!
! Copyright (c) 1983-2000 TAPP Consortium
! Copyright (c) 2000-2013,2016 Yoshihide Yoshimoto
!

module occup_num
contains
  !
  ! M. Methfessel and A.T. Paxton, PRB 40 (1989) 3616
  ! G. Kresse and J. Furthmueller, Computational Materials Science 6 (1996) 15
  !
  ! order = 1
  !
  real(kind=8) function wmethpax(x)
    implicit none
    real(kind=8),intent(in):: x

    real(kind=8),parameter:: pi = 3.14159265358979323846d0

    wmethpax = 0.5d0*( 1.0d0-erf(x) ) &
         - 1.0d0/(4.0d0*sqrt(pi))*(2.0d0*x)*exp(-x*x)
    return
  end function wmethpax

  real(kind=8) function smethpax(x)
    implicit none
    real(kind=8),intent(in):: x

    real(kind=8),parameter:: pi = 3.14159265358979323846d0

    smethpax = - 0.5d0/(4.0d0*sqrt(pi))*(4.0d0*x*x-2.0d0)*exp(-x*x)
    return
  end function smethpax

  !
  ! Fermi Dirac distribution
  !
  real(kind=8) function wfrmdrc(x)
    implicit none
    real(kind=8),intent(in):: x

    wfrmdrc = 1.0d0/(exp(x)+1.0d0)
    return
  end function wfrmdrc

  real(kind=8) function sfrmdrc(x)
    implicit none
    real(kind=8),intent(in):: x
    real(kind=8):: f

    f = 1.0d0/(exp(x)+1.0d0)
    if (f.eq.0.0d0 .or. f.eq.1.0d0) then
       sfrmdrc = 0.0d0
    else
       sfrmdrc = - ( f*log(f) + (1.0d0-f)*log(1.0d0-f) )
    end if
    return
  end function sfrmdrc

  subroutine doss(nband,ee,rco,nkmesh,pe,znel,dspin,ef &
       ,ndx,ndy,ndz,nb1,nb2,nkr,nng,nsy,kz,nks,ncomp,nspin,ekbt &
       ,wgtk,entrop,ismode,mssg)
    use mpi
    implicit none
    integer:: nband, nkmesh
    integer:: ndx, ndy, ndz, nb1, nb2, nkr, nng, nks, nspin, ismode, mssg
    integer:: ncomp
    real(kind=8),intent(in):: ee(nband,nks,nspin), rco(nkr,nks)
    real(kind=8),intent(out):: pe(nkr,nb1:nb2,nspin)
    real(kind=8),intent(in):: znel, dspin, ekbt
    real(kind=8),intent(out):: wgtk(nb1:nb2,nks,nspin)
    real(kind=8):: ef(nspin)
    integer:: nsy(nkr),kz(3,nkr,nng)
    real(kind=8),intent(out):: entrop

    integer,parameter:: ncycl = 100
    real(kind=8),parameter:: eps = 1.0d-14

    real(kind=8):: zne(2)
    integer:: iks, is, ib, jr
    integer:: istat,myrank
    integer:: info(2)
    real(kind=8):: occ(2)
    integer:: mthd
    real(kind=8):: tentrp

    if ( (ndx.eq.-1).and.(ndy.eq.-1).and.(ndz.eq.-1) ) then
       mthd = 2
    else if ( (ndx.eq.0).and.(ndy.eq.0).and.(ndz.eq.0) ) then
       mthd = 1
    else
       mthd = 0
    end if
    if ( (mthd.ne.0) .and. (mthd.ne.1) .and. (mthd.ne.2) ) then
       write(6,'(1x,a,i2)') 'invalid method for BZ int.', mthd
       call MPI_ABORT(MPI_COMM_WORLD,1,istat)
       stop
    end if

    pe = 0.0d0
    do is=1,nspin
       do iks=1,nks
          do ib=nb1,nb2
             do jr=1,nkr
                pe(jr,ib,is) = pe(jr,ib,is) &
                     + ee(ib,iks,is)*rco(jr,iks)/dble(nkmesh)
             end do
          end do
       end do
    end do

    if (mssg.gt.0) then
       write(6,'(1x,''coeff. of cosine exp. of bands'')')
       do ib=nb1,nb2
          write(6,'(3x,''ib= '',i4)') ib
          if (nspin .eq. 1) then
             do jr=1,nkr
                write(6,'(4x,i4,1x,e15.7)') jr, pe(jr,ib,1)
             end do
          else if (nspin .eq. 2) then
             do jr=1,nkr
                write(6,'(4x,i4,1x,e15.7,1x,e15.7)') &
                     jr, pe(jr,ib,1), pe(jr,ib,2)
             end do
          else
             write(6,*) 'invalid value of nspin = ', nspin
             call MPI_ABORT(MPI_COMM_WORLD,1,istat)
             stop
          end if
       end do
    end if

    if (nks.eq.1 .and. nkr.gt.1) then
       write(6,*)'ERROR NKR,NKS=',nkr,nks
       call MPI_ABORT(MPI_COMM_WORLD,1,istat)
       stop 'DOSS'
    end if

    if (mssg.gt.0) then
       if (mthd.eq.0) then
          if (nkr.eq.1) then
             write(6,'(1x,a,e12.4)') 'BZ integ. by Fermi dist: ekbt = ',ekbt
          else
             write(6,'(1x,a)') 'BZ integ. by cos exp. + tetrahedron'
          end if
       else if (mthd.eq.1) then
          write(6,'(1x,a,e12.4)') &
               'BZ integ. by Methfessel and Paxton: ekbt = ', ekbt
       else
          write(6,'(1x,a,e12.4)') &
               'BZ integ. by Fermi Dirac: ekbt = ', ekbt
       end if
       write(6,'(1x,a,1x,i2,1x,i2,1x,e12.4)') &
            'nspin,ismode,dspin =',nspin,ismode,dspin
    end if

    info = 0

    if (nspin .eq. 1) then
       zne(1) = znel*ncomp/2.0d0 - (nb1-1)
       if (mthd .eq. 0) then
          if (nkr .eq. 1) then
             call doswgtkpt(nband,nks,1,ee,nkr,rco,nb1,nb2,wgtk &
                  ,pe,nkmesh,zne(1),ef(1),occ(1),ekbt,entrop &
                  ,eps,ncycl,info(1),wfrmdrc,sfrmdrc)
             wgtk = 2.0d0/ncomp*wgtk
             entrop = 2.0d0/ncomp*entrop
          else
             call dos(pe(1,nb1,1),zne(1),ef(1) &
                  ,ndx,ndy,ndz,nb1,nb2,nkr,nng,nsy,kz,info(1),occ(1))
          end if
       else if (mthd .eq. 1) then
          call doswgtkpt(nband,nks,1,ee,nkr,rco,nb1,nb2,wgtk &
               ,pe,nkmesh,zne(1),ef(1),occ(1),ekbt,entrop &
               ,eps,ncycl,info(1),wmethpax,smethpax)
          wgtk = 2.0d0/ncomp*wgtk
          entrop = 2.0d0/ncomp*entrop
       else
          call doswgtkpt(nband,nks,1,ee,nkr,rco,nb1,nb2,wgtk &
               ,pe,nkmesh,zne(1),ef(1),occ(1),ekbt,entrop &
               ,eps,ncycl,info(1),wfrmdrc,sfrmdrc)
          wgtk = 2.0d0/ncomp*wgtk
          entrop = 2.0d0/ncomp*entrop
       end if
       pe(:,:,1) = 2.0d0/ncomp*pe(:,:,1)
    else if (nspin .eq. 2) then
       if (ismode .eq. 1) then
          zne(1) = (znel-2.0d0*(nb1-1))/2.0d0 + dspin/2.0d0
          zne(2) = (znel-2.0d0*(nb1-1))/2.0d0 - dspin/2.0d0
          entrop = 0.0d0
          do is=1,2
             if (mthd .eq. 0) then
                if (nkr .eq. 1) then
                   call doswgtkpt(nband,nks,1,ee(1,1,is),nkr,rco &
                        ,nb1,nb2,wgtk(nb1,1,is),pe(1,nb1,is) &
                        ,nkmesh,zne(is),ef(is),occ(is),ekbt,tentrp &
                        ,eps,ncycl,info(is),wfrmdrc,sfrmdrc)
                   entrop = entrop + tentrp
                else 
                   call dos(pe(1,nb1,is),zne(is),ef(is) &
                        ,ndx,ndy,ndz,nb1,nb2,nkr,nng,nsy,kz,info(is),occ(is))
                end if
             else if (mthd .eq. 1) then
                call doswgtkpt(nband,nks,1,ee(1,1,is),nkr,rco &
                     ,nb1,nb2,wgtk(nb1,1,is),pe(1,nb1,is) &
                     ,nkmesh,zne(is),ef(is),occ(is),ekbt,tentrp &
                     ,eps,ncycl,info(is),wmethpax,smethpax)
                entrop = entrop + tentrp
             else
                call doswgtkpt(nband,nks,1,ee(1,1,is),nkr,rco &
                     ,nb1,nb2,wgtk(nb1,1,is),pe(1,nb1,is) &
                     ,nkmesh,zne(is),ef(is),occ(is),ekbt,tentrp &
                     ,eps,ncycl,info(is),wfrmdrc,sfrmdrc)
                entrop = entrop + tentrp
             end if
          end do
       else if (ismode .eq. 2) then
          zne(1) = znel-2.0d0*(nb1-1)
          if (mthd .eq. 0) then
             if (nkr .eq. 1) then
                call doswgtkpt(nband,nks,nspin,ee,nkr,rco,nb1,nb2,wgtk &
                     ,pe,nkmesh,zne(1),ef(1),occ(1),ekbt,entrop &
                     ,eps,ncycl,info(1),wfrmdrc,sfrmdrc)
             else
                call dos(pe,zne(1),ef(1) &
                     ,ndx,ndy,ndz,1,2*(nb2-nb1+1),nkr,nng,nsy,kz &
                     ,info(1),occ(1))
             end if
          else if (mthd .eq. 1) then
             call doswgtkpt(nband,nks,nspin,ee,nkr,rco,nb1,nb2,wgtk &
                  ,pe,nkmesh,zne(1),ef(1),occ(1),ekbt,entrop &
                  ,eps,ncycl,info(1),wmethpax,smethpax)
          else
             call doswgtkpt(nband,nks,nspin,ee,nkr,rco,nb1,nb2,wgtk &
                  ,pe,nkmesh,zne(1),ef(1),occ(1),ekbt,entrop &
                  ,eps,ncycl,info(1),wfrmdrc,sfrmdrc)
          end if
          ef(2) = ef(1)
       else
          write(6,*) 'invalid value in ismode = ',ismode
          call MPI_ABORT(MPI_COMM_WORLD,1,istat)
          stop
       end if
    else
       write(6,*) 'invalid value in nspin = ',nspin
       call MPI_ABORT(MPI_COMM_WORLD,1,istat)
       stop
    end if
    if (maxval(info) .eq. 0) then
       if (mssg.gt.0) then
          write(6,'(1x,a)') ' EF CONVERGED'
          if ( (nspin.eq.1) .or. (nspin.eq.2 .and. ismode.eq.2) ) then
             write(6,'(1x,a,e26.16,a,e26.16)') 'EF=',ef(1),' CHARGE=',occ(1)
          else
             write(6,'(1x,a,e26.16,1x,e26.16)') 'EF=',ef(1),ef(2)
             write(6,'(1x,a,e26.16,1x,e26.16)') 'CHARGE=',occ(1),occ(2)
          end if
       end if
    else
       call MPI_COMM_RANK(MPI_COMM_WORLD,myrank,istat)
       write(6,'(1x,a)') ' EF IS NOT CONVERGED'
       write(6,'(1x,a,3(1x,i3))') 'myrank, info = ', myrank, info(1), info(2)
       write(6,'(/1x,a)') 'eigen energy'
       if (myrank.eq.0) then
          do is=1,nspin
             do iks=1,nks
                write(6,'(1x,a,1x,i4,1x,i2)') 'iks,is =',iks,is
                write(6,'(5(1x,e14.7))') (ee(ib,iks,is),ib=nb1,nb2)
             end do
             write(6,*)
          end do
       end if
       call MPI_FINALIZE(istat)
       stop
    end if
    if (mssg.gt.0) then
       do ib=nb1,nb2
          write(6,'(1x,a,i4)') 'IB=',ib
          if (nspin .eq. 1) then 
             do  jr=1,nkr
                write(6,'(1x,a,i4,3x,a,e14.6)') 'K=', jr, 'PE=',pe(jr,ib,1)
             end do
          else
             do jr=1,nkr
                write(6,'(1x,a,i4,3x,a,e14.6,1x,e14.6)') &
                     'K=', jr, 'PE=',pe(jr,ib,1),pe(jr,ib,2)
             end do
          end if
       end do
    end if

    if (mthd .eq. 0 .and. nkr .ne. 1) then
       do is=1,nspin
          call weightk(nb1, nb2, nkr, nks, pe(1,nb1,is), rco, nsy &
               ,nband, ee(1,1,is), 0.0d0, wgtk(nb1,1,is), mssg)
       end do
       entrop = 0.0d0
    else
       if (mssg.gt.0) then
          write(6,'(1x,a,e14.6)') 'Entropy = ', entrop
       end if
    end if
    return
  end subroutine doss

  subroutine dos(pe,zne,ef,ndx,ndy,ndz,nb1,nb2,nkr,nng,nsy,kz,info,occ)
    implicit none
    integer nb1,nb2,nkr,ndx,ndy,ndz,nng
    real(kind=8) zne,ef
    real(kind=8) pe(nkr,nb1:nb2)
    integer nsy(nkr),kz(3,nkr,nng)
    integer info
    real(kind=8) occ
    !
    integer,parameter:: mcycl = 100
    real(kind=8),parameter:: eps = 1.0d-9
    !
    real(kind=8),dimension(:,:,:,:):: emesh1,emesh2
    real(kind=8),dimension(:,:):: vp1,vp2
    allocatable emesh1,emesh2,vp1,vp2
    integer ierr
    !
    allocate(emesh1(-ndx:ndx,-ndy:ndy,0:ndz,nb1:nb2), &
         emesh2(-ndx*2:ndx*2,-ndy*2:ndy*2,0:ndz*2,nb1:nb2), &
         vp1(nkr,nb1:nb2),vp2(nkr,nb1:nb2), &
         stat=ierr)
    call chkalloc('DOS','emesh',ierr)
    !
    call edef(pe,emesh1,emesh2,nsy,kz &
         ,ndx,ndy,ndz,nb1,nb2,nkr,nng)
    !
    call fermi(emesh1,emesh2,ef &
         ,ndx,ndy,ndz,zne,nb1,nb2,mcycl,eps,info,occ)
    if (info .ne. 0) return
    !
    call vinteg(emesh1,kz(1,1,1),vp1,ef &
         ,ndx,ndy,ndz,nkr,nb1,nb2)
    call vinteg(emesh2,kz(1,1,1),vp2,ef &
         ,ndx*2,ndy*2,ndz*2,nkr,nb1,nb2)
    call romb(pe,vp1,vp2,nkr,nb1,nb2)
    !
    deallocate(emesh1,emesh2,vp1,vp2)
    !
    return
  end subroutine dos

  subroutine store_pe(nband,nspin,ee,nb1,nb2,mgs)
    use iounit
    use meshgrid
    implicit none
    integer:: nband, nspin, nb1, nb2
    type (meshgrid_struct):: mgs
    real(kind=8),intent(in):: ee(nband,mgs%nks,nspin)

    real(kind=8),allocatable:: pe(:,:,:)

    integer:: nks,nkr,nng,nkmesh
!soption contarea(rco,nsy,kz)
    real(kind=8),pointer:: rco(:,:)
    integer,pointer:: nsy(:),kz(:,:,:)

    integer:: iks, is, ib, jr, ierr

    nks = mgs%nks
    nkr = mgs%nkr
    nng = mgs%nng
    nkmesh = mgs%nkmesh
    rco => mgs%rco
    nsy => mgs%nsy
    kz => mgs%kz

    allocate(pe(nkr,nb1:nb2,nspin),stat=ierr)
    call chkalloc('store_pe','tmp',ierr)

    pe = 0.0d0
    do is=1,nspin
       do iks=1,nks
          do ib=nb1,nb2
             do jr=1,nkr
                pe(jr,ib,is) = pe(jr,ib,is) &
                     + ee(ib,iks,is)*rco(jr,iks)/dble(nkmesh)
             end do
          end do
       end do
    end do
    call xrewind(NFPE)
    write(NFPE) nb1,nb2,nkr,nng
    write(NFPE) pe,kz,nsy
    deallocate(pe)

    return
  end subroutine store_pe

  subroutine weightk(nb1, nb2, nkr, nks, pe, rco, nsy, &
       nband, ee, de, wgtk, mssg)
    implicit none
    integer, intent(in):: nb1, nb2, nkr, nks
    real(kind=8),intent(in):: pe(nkr, nb1:nb2), rco(nkr, nks)
    integer,intent(in):: nsy(nkr)
    integer,intent(in):: nband
    real(kind=8),intent(in):: ee(nband, nks), de
    real(kind=8), intent(out)::  wgtk(nb1:nb2, nks)
    integer,intent(in):: mssg

    integer:: iks, ib, jb, ikr, mb1, mb2
    real(kind=8):: tmp, diff
    real(kind=8), parameter:: pi = 3.14159265358979323846d0
    real(kind=8),allocatable:: alph(:), twgtk(:, :)
    integer:: ierr, count

    allocate(twgtk(nb1:nb2, nks), alph(nb1:nb2),stat=ierr)
    call chkalloc('weightk','twgtk, alph',ierr)

    twgtk = 0.0d0
    do iks=1,nks
       do ib=nb1, nb2
          do ikr=1, nkr
             twgtk(ib, iks) = twgtk(ib, iks) &
                  + pe(ikr,ib)*rco(ikr,iks)*nsy(ikr)
          end do
       end do
    end do

    if ( (de .eq. 0.0d0) .or. (nb1 .eq. nb2) ) then
       wgtk = twgtk
       deallocate(twgtk, alph)
       return
    end if

    if (nks .eq. 1) then
       if (nkr .eq. 1) then
          if (mssg .gt. 0) then
             write(6,*) 'weightk: doss should have fixed the weight.'
             write(6,*) 'weightk: skipping'
          end if
          wgtk = twgtk
          deallocate(twgtk, alph)
          return
       else
          write(6,*) 'error: nkr, nks = ', nkr, nks
          deallocate(twgtk, alph)
          stop
       end if
    end if

    ! detect the degenerated eigen values and fix the weight.

    count = 0
    wgtk = 0.0d0
    do iks=1, nks
       do ib=nb1, nb2
          alph(ib) = 1.0d0
          tmp = alph(ib)
          mb1 = -1
          do jb=ib-1, nb1, -1
             diff = abs(ee(jb, iks) - ee(ib, iks))
             if (diff .lt. de) then
                alph(jb) = 0.5d0*( cos(pi*diff/de) + 1.0d0 )
                tmp = tmp + alph(jb)
             else
                mb1 = jb + 1
                exit
             end if
          end do
          if (mb1 .eq. -1) then
             mb1 = nb1
          end if
          mb2 = -1
          do jb=ib+1, nb2, 1
             diff = abs(ee(jb, iks) - ee(ib, iks))
             if (diff .lt. de) then
                alph(jb) = 0.5d0*( cos(pi*diff/de) + 1.0d0 )
                tmp = tmp + alph(jb)
             else
                mb2 = jb - 1
                exit
             end if
          end do
          if (mb2 .eq. -1) then
             mb2 = nb2
          end if
          do jb=mb1, mb2
             wgtk(jb, iks) = wgtk(jb, iks) + alph(jb)/tmp*twgtk(ib, iks)
          end do
          if (mb1 .lt. mb2) then
             count = count + 1
          end if
       end do
    end do
    if (mssg .gt. 0) then
       write(6,'(1x,a,i4,a)') 'weightk: fixed ', count, ' weights'
    end if
    deallocate(twgtk, alph)
    return
  end subroutine weightk

  !
  ! general routine for weighted k-point type DOS calculation.
  !
  subroutine doswgtkpt(nband,nks,nspin,ee,nkr,rco,nb1,nb2,wgtk &
       ,pe,nkmesh,zne,ef,occ,ekbt,entrop,eps,ncycl,info,wfnc,sfnc)
    implicit none
    integer:: nband, nb1, nb2, nkr, nks, nspin,nkmesh
    real(kind=8),intent(in):: ee(nband,nks,nspin), rco(nkr,nks)
    real(kind=8),intent(out):: pe(nkr,nb1:nb2,nspin)
    real(kind=8),intent(in):: zne, ekbt, eps
    real(kind=8),intent(out):: wgtk(nb1:nb2,nks,nspin),entrop
    real(kind=8),intent(out):: ef,occ
    integer:: ncycl, info
    interface
       real(kind=8) function wfnc(x)
         real(kind=8),intent(in):: x
       end function wfnc
       real(kind=8) function sfnc(x)
         real(kind=8),intent(in):: x
       end function sfnc
    end interface

    integer,parameter:: nefs = 20
    real(kind=8):: occv(0:nefs)
    integer:: pos(nefs)
    real(kind=8):: nego(2),cef(2)

    integer:: iks,ic,ib,is,ief,nfnd
    real(kind=8):: ef1,ef2,f,ff,x,w,einf,esup,efv,delta

    ef1=minval(ee)-(-log(eps))*ekbt
    ef2=maxval(ee)+(-log(eps))*ekbt

    info = 1
    do ic=1,ncycl
       ef = 0.5d0*(ef1+ef2)
       if (ef.eq.ef1 .or. ef.eq.ef2) then
          info = 0
          exit
       end if
       occ = 0.0d0
       do is=1,nspin
          do iks=1,nks
             do ib=nb1,nb2
                x = (ee(ib,iks,is) - ef)/ekbt
                f = wfnc(x)
                occ = occ + f*rco(1,iks)
             end do
          end do
       end do
       occ = occ/nkmesh
       if (occ-zne.gt.eps) then
          ef2 = ef
       else if (occ-zne.lt.-eps) then
          ef1 = ef
       else
          info = 0
          exit
       end if
    end do
    if (info .ne. 0) then
       return
    end if

    !
    ! Seek for another solution.
    ! Since the weight function is not monotonous,
    ! there may be another solution, especially for gapful systems.
    !

    einf = -huge(einf)
    esup = huge(esup)
    do is=1,nspin
       do iks=1,nks
          do ib=nb1,nb2
             if (ee(ib,iks,is).lt.ef) then
                einf = max(einf,ee(ib,iks,is))
             else
                esup = min(esup,ee(ib,iks,is))
             end if
          end do
       end do
    end do

    if ( (esup-einf).lt.ekbt ) then
       delta = (ekbt-esup+einf)*0.5d0
       esup = esup + delta
       einf = einf - delta
    else
       esup = esup + 0.5d0*ekbt
       einf = einf - 0.5d0*ekbt 
    end if
    
    delta = (esup-einf)/nefs
    do ief=0,nefs
       efv = ief*delta + einf
       occv(ief) = 0.0d0
       do is=1,nspin
          do iks=1,nks
             do ib=nb1,nb2
                x = (ee(ib,iks,is) - efv)/ekbt
                f = wfnc(x)
                occv(ief) = occv(ief) + f*rco(1,iks)
             end do
          end do
       end do
       occv(ief) = occv(ief)/nkmesh
    end do
    nfnd = 0
    do ief=0,nefs-1
       if (occv(ief).le.zne .and. occv(ief+1).ge.zne) then
          nfnd = nfnd + 1
          pos(nfnd) = ief
       end if
    end do
    if (nfnd.le.0) then
       info = 2
       return
    else if (nfnd.gt.1) then
       ief = pos(1)
       ef1 = ief*delta + einf
       ef2 = (ief+1)*delta + einf

       info = 3
       do ic=1,ncycl
          ef = 0.5d0*(ef1+ef2)
          if (ef.eq.ef1 .or. ef.eq.ef2) then
             info = 0
             exit
          end if
          occ = 0.0d0
          nego(1) = 0.0d0
          do is=1,nspin
             do iks=1,nks
                do ib=nb1,nb2
                   x = (ee(ib,iks,is) - ef)/ekbt
                   f = wfnc(x)
                   occ = occ + f*rco(1,iks)
                   nego(1) = nego(1) + min(0.0d0,f*rco(1,iks))
                end do
             end do
          end do
          occ = occ/nkmesh
          if (occ-zne.gt.eps) then
             ef2 = ef
          else if (occ-zne.lt.-eps) then
             ef1 = ef
          else
             info = 0
             exit
          end if
       end do
       if (info .ne. 0) then
          return
       end if
       cef(1) = ef
       nego(1) = abs(nego(1))

       ief = pos(nfnd)
       ef1 = ief*delta + einf
       ef2 = (ief+1)*delta + einf

       info = 3
       do ic=1,ncycl
          ef = 0.5d0*(ef1+ef2)
          if (ef.eq.ef1 .or. ef.eq.ef2) then
             info = 0
             exit
          end if
          occ = 0.0d0
          nego(2) = 0.0d0
          do is=1,nspin
             do iks=1,nks
                do ib=nb1,nb2
                   x = (ee(ib,iks,is) - ef)/ekbt
                   f = wfnc(x)
                   occ = occ + f*rco(1,iks)
                   nego(2) = nego(2) + min(0.0d0,f*rco(1,iks))
                end do
             end do
          end do
          occ = occ/nkmesh
          if (occ-zne.gt.eps) then
             ef2 = ef
          else if (occ-zne.lt.-eps) then
             ef1 = ef
          else
             info = 0
             exit
          end if
       end do
       if (info .ne. 0) then
          return
       end if
       cef(2) = ef
       nego(2) = abs(nego(2))

       if (nego(1).lt.nego(2)) then
          ef = cef(1)
       else
          ef = cef(2)
       end if
    end if

    entrop = 0.0d0
    do is=1,nspin
       do iks=1,nks
          do ib=nb1,nb2
             x = (ee(ib,iks,is) - ef)/ekbt
             f = wfnc(x)
             ff = sfnc(x)
             wgtk(ib,iks,is) = f*rco(1,iks)
             entrop = entrop + ekbt*ff*rco(1,iks)
          end do
       end do
    end do
    entrop = entrop/nkmesh

    pe = 0.0d0
    do is=1,nspin
       do ib=nb1,nb2
          w = 0.0d0
          do iks=1,nks
             pe(1,ib,is) = pe(1,ib,is) + wgtk(ib,iks,is)
             w = w + rco(1,iks)
          end do
          pe(1,ib,is) = pe(1,ib,is)/w
       end do
    end do

  end subroutine doswgtkpt
end module occup_num
