/*
  Copyright (c) 2012-2013 Yoshihide Yoshimoto
 */

#include <cuda.h>
#include <cutil.h>
#include <cufft.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "exchange_kern_cuda.h"

#undef DEBUG_MODE

#ifdef DEBUG_MODE
#define CHECKSTAT(place,stat) { printf(place,stat); }
#else
#define CHECKSTAT(place,stat)
#endif

static cufftHandle plHandleF, plHandleB;
static cudaDeviceProp deviceProp;
static int IfCudaDevInit = 0;

static double *SRmobj;
static double *sumres;
static const int cores4sm = 32;

static int ifgam = 0;

static void cu_fft_setup_() {
    int deviceCount;

    if (IfCudaDevInit == 1) {
        return;
    }

    cuInit(0);

    CUDA_SAFE_CALL(cudaGetDeviceCount(&deviceCount));
    if (deviceCount>0) {
        cudaGetDeviceProperties(&deviceProp, 0);
#ifdef DEBUG_MODE
        printf("device name = %s\n",deviceProp.name);
#endif
        CUDA_SAFE_CALL(cudaSetDevice(0));
        IfCudaDevInit = 1;
    } else {
        fprintf(stderr,"no GPU device\n");
    }
}

static void cu_fft_create_(int *frh,int *nx,int *ny,int *nz) {
    cufftResult stat;

    if (*frh == 1) {
        stat = cufftPlan3d(&plHandleF, *nz, *ny, *nx, CUFFT_D2Z);
        CHECKSTAT("cufftplan3d %d\n",stat);
        stat = cufftPlan3d(&plHandleB, *nz, *ny, *nx, CUFFT_Z2D);
        CHECKSTAT("cufftplan3d %d\n",stat);
    } else {
        stat = cufftPlan3d(&plHandleF, *nz, *ny, *nx, CUFFT_Z2Z);
        CHECKSTAT("cufftplan3d %d\n",stat);
        stat = cufftPlan3d(&plHandleB, *nz, *ny, *nx, CUFFT_Z2Z);
        CHECKSTAT("cufftplan3d %d\n",stat);
    }
}

static void cu_fft_destroy_() {
    cufftResult ret;

    ret = cufftDestroy(plHandleF);
    ret = cufftDestroy(plHandleB);

}

static void cu_fft_teardown_() {

}

__global__ void kernel_zero_d(double *x) {
    int gid ;
    gid = blockDim.x*blockIdx.x + threadIdx.x;
    x[gid] = 0.0;
}

__global__ void kernel_zero_d2(double2 *x) {
    int gid ;
    gid = blockDim.x*blockIdx.x + threadIdx.x;
    x[gid].x = 0.0;
    x[gid].y = 0.0;
}

__global__ void kernel_mul_zc_z(double2 *x, double2 *y, double2 *z) {
    int gid ;
    double2 tz;
    gid = blockDim.x*blockIdx.x + threadIdx.x;
    tz.x = x[gid].x*y[gid].x + x[gid].y*y[gid].y;
    tz.y = x[gid].x*y[gid].y - x[gid].y*y[gid].x;
    z[gid] = tz;
}

__global__ void kernel_mul_d2(double2 *x, double2 *y, double2 *z) {
    int gid ;
    double2 tz;
    gid = blockDim.x*blockIdx.x + threadIdx.x;
    tz.x = x[gid].x*y[gid].x;
    tz.y = x[gid].y*y[gid].y;
    z[gid] = tz;
}

__global__ void kernel_zdscal_v(double *x, double2 *y) {
    int gid ;
    gid = blockDim.x*blockIdx.x + threadIdx.x;
    y[gid].x = x[gid]*y[gid].x;
    y[gid].y = x[gid]*y[gid].y;
}

__global__ void kernel_mul_add_z_z_scal(
    double2 *x,  double2 *y, double2 *z, double f) {
    int gid ;
    double2 tz;
    gid = blockDim.x*blockIdx.x + threadIdx.x;
    tz.x = x[gid].x*y[gid].x - x[gid].y*y[gid].y;
    tz.y = x[gid].x*y[gid].y + x[gid].y*y[gid].x;
    z[gid].x = z[gid].x + f*tz.x;
    z[gid].y = z[gid].y + f*tz.y;
}

__global__ void kernel_mul_add_d2_scal(
    double2 *x,  double2 *y, double2 *z, double f) {
    int gid ;
    double2 tz;
    gid = blockDim.x*blockIdx.x + threadIdx.x;
    tz.x = x[gid].x*y[gid].x;
    tz.y = x[gid].y*y[gid].y;
    z[gid].x = z[gid].x + f*tz.x;
    z[gid].y = z[gid].y + f*tz.y;
}

__global__ void kernel_zdscal3_v(
    double *x, double2 *y, double2 *z) {
    int gid;
    gid = blockDim.x*blockIdx.x + threadIdx.x;
    z[gid].x = x[gid]*y[gid].x;
    z[gid].y = x[gid]*y[gid].y;
}

__global__ void kernel_sumfzxy(
    double2 *x, double2 *y, double2 *z,
    double f, int len, double* res, int rbase) {
    extern __shared__ double scr[];
    int ng, i;
    double ar, ai, sum;
    ng = gridDim.x*blockDim.x;
    sum = 0.0;
    i = blockDim.x*blockIdx.x + threadIdx.x;
    while (i < len) {
        ar = x[i].x*y[i].x - x[i].y*y[i].y;
        ai = x[i].x*y[i].y + x[i].y*y[i].x;
        sum = sum + f*(z[i].x*ar + z[i].y*ai);
        i += ng;
    }
    i = threadIdx.x;
    scr[i] = sum;
    if (i == 0) {
        sum = 0.0;
        for (i=0; i<blockDim.x; i++) {
            sum = sum + scr[i];
        }
        res[blockIdx.x+rbase] += sum;
    }
}

static cuDoubleComplex *ztmp, *ztmp2, *pxi;
static cuDoubleComplex **pxj;
static cuDoubleComplex *vsum;
static double *CKmobj;
static double *CKSmobj[6];
static double *clkern;
static double *clkerns;
static cuDoubleComplex *tvxp;
static int nlx,nly,nlz,nlxyz;
static int nwx,nwy,nwz,nwx2,nwy2,nwz2,nwxyz;
static int infwx;
static int niw,njw;

extern "C"
void get_exchange_nlxyz_(int *nlxo, int *nlyo, int *nlzo) {
    *nlxo = nlx;
    *nlyo = nly;
    *nlzo = nlz;
}

extern "C"
void init_exchange_(int *ifgami, int *niwi, int *njwi, int* nwxi, int *nwyi, int *nwzi) {
    int ib;

    ifgam = *ifgami;
    niw = *niwi;
    njw = *njwi;
    nwx = *nwxi;
    nwy = *nwyi;
    nwz = *nwzi;
    nwx2 = nwx*2;
    nwy2 = nwy*2;
    nwz2 = nwz*2;
    nwxyz = nwx2*nwy2*nwz2;
    if (ifgam == 1) {
        infwx = 0;
        nlx = nwx + 1;
    } else {
        infwx = -nwx+1;
        nlx = nwx2;
    }
    nly = nwy2;
    nlz = nwz2;
    nlxyz = nlx*nly*nlz;

    cu_fft_setup_();

    cu_fft_create_(&ifgam, &nwx2, &nwy2, &nwz2);

    CUDA_SAFE_CALL(cudaMalloc((void **)&pxi,nlxyz*sizeof(cuDoubleComplex)));
    pxj = (cuDoubleComplex **)malloc(sizeof(cuDoubleComplex *)*njw);
    tvxp = (cuDoubleComplex *)malloc(niw*nlxyz*sizeof(cuDoubleComplex));
    clkern = (double *)malloc(nlxyz*sizeof(double));
    clkerns = (double *)malloc(nlxyz*sizeof(double)*6);
    sumres = (double *)malloc(sizeof(double)*deviceProp.multiProcessorCount*7);

    for (ib=0; ib<njw; ib++) {
        CUDA_SAFE_CALL(cudaMalloc((void **)&(pxj[ib]), nlxyz*sizeof(cuDoubleComplex)));
    }

    CUDA_SAFE_CALL(cudaMalloc((void **)&SRmobj, sizeof(double)*deviceProp.multiProcessorCount*7));
    CUDA_SAFE_CALL(cudaMalloc((void **)&ztmp, nlxyz*sizeof(cuDoubleComplex)));
    CUDA_SAFE_CALL(cudaMalloc((void **)&ztmp2, nlxyz*sizeof(cuDoubleComplex)));
    CUDA_SAFE_CALL(cudaMalloc((void **)&vsum, nlxyz*sizeof(cuDoubleComplex)));
    CUDA_SAFE_CALL(cudaMalloc((void **)&CKmobj, nlxyz*sizeof(double)));
    for (ib=0; ib<6; ib++) {
        CUDA_SAFE_CALL(cudaMalloc((void **)&(CKSmobj[ib]),nlxyz*sizeof(double)))
    }

    CHECKSTAT("allocate %d\n", 1);
}

extern "C"
void setup_exchange_clkern_(const int *nkmesh, const double bb[3][3], const double q[3], const double *rcp, const double *rdp) {

    const double pi = 3.14159265358979323846;

    int ix,iy,iz,i1,i2,i3,ig;
    double aqi[3], g[3];
    double vb,rc,rd,gsq,ga,dsq;
    double epsilon;

    memset(clkern, 0, sizeof(double)*nlxyz);

    vb = bb[0][0]*(bb[1][1]*bb[2][2]-bb[1][2]*bb[2][1]) 
        +bb[0][1]*(bb[1][2]*bb[2][0]-bb[1][0]*bb[2][2]) 
        +bb[0][2]*(bb[1][0]*bb[2][1]-bb[1][1]*bb[2][0]);
    vb = fabs(vb);

    rc = *rcp;
    rd = *rdp;
    dsq = pow(pi/(rd-rc), 2.0);
    epsilon = 1.0e-6*pow(vb/(*nkmesh), 1.0/3.0);

#pragma omp parallel for private(ix,iy,iz,i1,i2,i3,aqi,ig,g,gsq,ga)
    for(iz=-nwz+1; iz<=nwz-1; iz++) {
        i3 = (iz+nwz2)%nwz2;
        aqi[2] = q[2] + iz;
        for(iy=-nwy+1; iy<=nwy-1; iy++) {
            i2 = (iy+nwy2)%nwy2;
            aqi[1] = q[1] + iy;
            for(ix=infwx; ix<=nwx-1; ix++) {
                i1 = (ix+nwx2)%nwx2;
                aqi[0] = q[0] + ix;

                ig = i1 + nlx*i2 + nlx*nly*i3;

                g[0] = bb[0][0]*aqi[0]+bb[1][0]*aqi[1]+bb[2][0]*aqi[2];
                g[1] = bb[0][1]*aqi[0]+bb[1][1]*aqi[1]+bb[2][1]*aqi[2];
                g[2] = bb[0][2]*aqi[0]+bb[1][2]*aqi[1]+bb[2][2]*aqi[2];

                gsq = g[0]*g[0] + g[1]*g[1] + g[2]*g[2];
                ga = sqrt(gsq);

                if (ga < epsilon) {
                    clkern[ig] = pi*(rc*rc+rd*rd) - 4.0/pi*(rd-rc)*(rd-rc);
                } else {
                    clkern[ig] = 4.0*pi/gsq + 
                        2.0*pi*( -1.0/gsq + 1.0/(gsq-dsq) )
                        *( cos(ga*rc) + cos(ga*rd) );
                }
                clkern[ig] = clkern[ig]/nwxyz; /* include FFT_FORWARD factor */
            }
        }
    }
    CUDA_SAFE_CALL(cudaMemcpy(CKmobj, clkern, nlxyz*sizeof(double), cudaMemcpyHostToDevice));

    CHECKSTAT("initclkern %d\n",0);
}

extern "C"
void setup_exchange_clkerns_(const int *nkmesh, const double bb[3][3], const double q[3], const double *rcp, const double *rdp) {

    const double pi = 3.14159265358979323846;

    int ix,iy,iz,i1,i2,i3,ig;
    double aqi[3], g[3];
    double vb,rc,rd,gsq,ga,dsq,fac;
    double epsilon;

    memset(clkerns, 0, sizeof(double)*nlxyz*6);

    vb = bb[0][0]*(bb[1][1]*bb[2][2]-bb[1][2]*bb[2][1]) 
        +bb[0][1]*(bb[1][2]*bb[2][0]-bb[1][0]*bb[2][2]) 
        +bb[0][2]*(bb[1][0]*bb[2][1]-bb[1][1]*bb[2][0]);
    vb = fabs(vb);

    rc = *rcp;
    rd = *rdp;
    dsq = pow(pi/(rd-rc), 2.0);
    epsilon = 1.0e-6*pow(vb/(*nkmesh), 1.0/3.0);

#pragma omp parallel for private(ix,iy,iz,i1,i2,i3,aqi,ig,g,gsq,ga,fac)
    for(iz=-nwz+1; iz<=nwz-1; iz++) {
        i3 = (iz+nwz2)%nwz2;
        aqi[2] = q[2] + iz;
        for(iy=-nwy+1; iy<=nwy-1; iy++) {
            i2 = (iy+nwy2)%nwy2;
            aqi[1] = q[1] + iy;
            for(ix=infwx; ix<=nwx-1; ix++) {
                i1 = (ix+nwx2)%nwx2;
                aqi[0] = q[0] + ix;

                ig = i1 + nlx*i2 + nlx*nly*i3;

                g[0] = bb[0][0]*aqi[0]+bb[1][0]*aqi[1]+bb[2][0]*aqi[2];
                g[1] = bb[0][1]*aqi[0]+bb[1][1]*aqi[1]+bb[2][1]*aqi[2];
                g[2] = bb[0][2]*aqi[0]+bb[1][2]*aqi[1]+bb[2][2]*aqi[2];

                gsq = g[0]*g[0] + g[1]*g[1] + g[2]*g[2];
                ga = sqrt(gsq);

                if (ga < epsilon) {
                    clkerns[ig] = 0.0;
                    clkerns[ig+nlxyz] = 0.0;
                    clkerns[ig+2*nlxyz] = 0.0;
                    clkerns[ig+3*nlxyz] = 0.0;
                    clkerns[ig+4*nlxyz] = 0.0;
                    clkerns[ig+5*nlxyz] = 0.0;
                } else {
                    fac = -8.0*pi/gsq +
                        4.0*pi*( 1.0/gsq - gsq/((gsq-dsq)*(gsq-dsq)) )
                        *( cos(ga*rc) + cos(ga*rd) ) +
                        2.0*pi*( 1.0/gsq - 1.0/(gsq-dsq) )
                        *( ga*rc*sin(ga*rc) + ga*rd*sin(ga*rd) );

                    fac = fac/nwxyz; /* include FFT_FORWARD factor */
                    g[0] = g[0]/ga;
                    g[1] = g[1]/ga;
                    g[2] = g[2]/ga;

                    clkerns[ig] = fac*g[0]*g[0];
                    clkerns[ig+nlxyz] = fac*g[1]*g[1];
                    clkerns[ig+2*nlxyz] = fac*g[2]*g[2];
                    clkerns[ig+3*nlxyz] = fac*2.0*g[1]*g[2];
                    clkerns[ig+4*nlxyz] = fac*2.0*g[2]*g[0];
                    clkerns[ig+5*nlxyz] = fac*2.0*g[0]*g[1];
                }
            }
        }
    }
    for (ig=0; ig<6; ig++) {
        CUDA_SAFE_CALL(cudaMemcpy(CKSmobj[ig], clkerns+ig*nlxyz, nlxyz*sizeof(double), cudaMemcpyHostToDevice));
    }

    CHECKSTAT("initclkerns %d\n",0);
}

extern "C"
void teardown_exchange_() {
    int ib;

    CUDA_SAFE_CALL(cudaFree(SRmobj));
    CUDA_SAFE_CALL(cudaFree(ztmp));
    CUDA_SAFE_CALL(cudaFree(ztmp2));
    CUDA_SAFE_CALL(cudaFree(vsum));
    CUDA_SAFE_CALL(cudaFree(CKmobj));
    for (ib=0;ib<6;ib++) {
        CUDA_SAFE_CALL(cudaFree(CKSmobj[ib]));
    }

    CUDA_SAFE_CALL(cudaFree(pxi));
    for (ib=0; ib<njw; ib++) {
        CUDA_SAFE_CALL(cudaFree(pxj[ib]));
    }
    free(sumres);
    free(clkern);
    free(tvxp);
    free(pxj);

    cu_fft_destroy_();

    cu_fft_teardown_();
}

extern "C"
void deploy_pxj_kern_(int *mjwi, double_complex *wfj_in) {
    cufftResult stat;

    cuDoubleComplex *wf1;
    int mjw, ib;

    mjw = *mjwi;
    if (mjw > njw) {
        fprintf(stderr,"invalid mjw = %d\n", mjw);
        exit(1);
    }

    wf1 = (cuDoubleComplex *) wfj_in;
    for(ib=0; ib<mjw; ib++) {
        CUDA_SAFE_CALL(cudaMemcpy(ztmp, wf1, nlxyz*sizeof(cuDoubleComplex), cudaMemcpyHostToDevice));
        if (ifgam == 1) {
            stat = cufftExecZ2D(plHandleB, ztmp, (cufftDoubleReal *)pxj[ib]);
            CHECKSTAT("fftbw %d\n", stat);
        } else {
            stat = cufftExecZ2Z(plHandleB, ztmp, pxj[ib], CUFFT_INVERSE);
            CHECKSTAT("fftbw %d\n", stat);
        }
        wf1 = wf1 + nlxyz;
    }
}

extern "C"
void compute_exchange_(int *miwi, int *mjwi, double_complex *wfi_in, double_complex *wfj_in, double_complex *vxp_in, double *wgt) {
    cuDoubleComplex *wfi, *vxp;
    int ig;
    int ib,jb;

    double fac;

    cufftResult stat;

    cuDoubleComplex *wf1, *vxp1, *tvxp1;
    int miw, mjw;

    wfi = (cuDoubleComplex *) wfi_in;
    vxp = (cuDoubleComplex *) vxp_in;

    miw = *miwi;
    mjw = *mjwi;
    if (miw > niw || mjw > njw) {
        fprintf(stderr,"invalid miw, mjw = %d %d\n", miw, mjw);
        exit(1);
    }

    kernel_zero_d2 <<<nlx*nly, nlz>>> ((double2 *)ztmp);
    CUT_CHECK_ERROR("zero fail\n");
    kernel_zero_d2 <<<nlx*nly, nlz>>> ((double2 *)ztmp2);
    CUT_CHECK_ERROR("zero fail\n");

    wf1 = wfi;
    tvxp1 = tvxp;
    for(ib=0; ib<miw; ib++) {
        CUDA_SAFE_CALL(cudaMemcpy(ztmp, wf1, nlxyz*sizeof(cuDoubleComplex), cudaMemcpyHostToDevice));
        wf1 = wf1 + nlxyz;

        if (ifgam == 1) {
            stat = cufftExecZ2D(plHandleB, ztmp, (cufftDoubleReal *)pxi);
            CHECKSTAT("fft bw %d\n",stat);
        } else {
            stat = cufftExecZ2Z(plHandleB, ztmp, pxi, CUFFT_INVERSE);
            CHECKSTAT("fft bw %d\n",stat);
        }

        kernel_zero_d2<<<nlx*nly,nlz>>>((double2 *)vsum);
        CUT_CHECK_ERROR("zero fail\n");

        for(jb=0; jb<mjw; jb++) {
            if (ifgam == 1) {
                kernel_mul_d2<<<nlx*nly,nlz>>>((double2 *)(pxj[jb]),(double2 *)pxi,(double2 *)ztmp);
                CUT_CHECK_ERROR("mul_d2 fail\n");
            } else {
                kernel_mul_zc_z<<<nlx*nly,nlz>>>((double2 *)(pxj[jb]),(double2 *)pxi,(double2 *)ztmp);
                CUT_CHECK_ERROR("mul_zc_z fail\n");
            }

            /* */

            if (ifgam == 1) {
                stat = cufftExecD2Z(plHandleF, (cufftDoubleReal *)ztmp, ztmp2);
                CHECKSTAT("fft fw %d\n",stat);
            } else {
                stat = cufftExecZ2Z(plHandleF, ztmp, ztmp2, CUFFT_FORWARD);
                CHECKSTAT("fft fw %d\n",stat);
            }

            /* CKmobj includes FFT_FOWARD factor */

            kernel_zdscal_v<<<nlx*nly,nlz>>>(CKmobj,(double2 *)ztmp2);
            CUT_CHECK_ERROR("zdscal_v fail\n");

            /* */

            if (ifgam == 1) {
                stat = cufftExecZ2D(plHandleB, ztmp2, (cufftDoubleReal *)ztmp);
                CHECKSTAT("fft bw %d\n",stat);

                /* */

                kernel_mul_add_d2_scal<<<nlx*nly,nlz>>>((double2 *)ztmp,(double2 *)pxj[jb],(double2 *)vsum,wgt[jb]);
                CUT_CHECK_ERROR("mul_add_d2 fail");
            } else {
                stat = cufftExecZ2Z(plHandleB, ztmp2, ztmp, CUFFT_INVERSE);
                CHECKSTAT("fft bw %d\n",stat);

                /* */

                kernel_mul_add_z_z_scal<<<nlx*nly,nlz>>>((double2 *)ztmp,(double2 *)pxj[jb],(double2 *)vsum,wgt[jb]);
                CUT_CHECK_ERROR("mul_add_z_z fail");
            }
        }
        if (ifgam == 1) {
            stat = cufftExecD2Z(plHandleF, (cufftDoubleReal *)vsum, ztmp);
            CHECKSTAT("fft fw %d\n",stat);
        } else {
            stat = cufftExecZ2Z(plHandleF, vsum, ztmp, CUFFT_FORWARD);
            CHECKSTAT("fft fw %d\n",stat);
        }

        /* FFT_FORWARD factor will be applied in main memory */

        CUDA_SAFE_CALL(cudaMemcpy(tvxp1, ztmp, nlxyz*sizeof(cuDoubleComplex), cudaMemcpyDeviceToHost));
        tvxp1 = tvxp1 + nlxyz;
    }

    fac = 1.0/nwxyz; /* include FFT_FORWARD factor */
#pragma omp parallel for private(ig,vxp1,tvxp1)
    for(ib=0; ib<miw; ib++) {
        vxp1 = vxp + ib*nlxyz;
        tvxp1 = tvxp + ib*nlxyz;
        for (ig=0; ig<nlxyz; ig++) {
            vxp1[ig] = make_cuDoubleComplex(
                cuCreal(vxp1[ig]) - fac*cuCreal(tvxp1[ig]),
                cuCimag(vxp1[ig]) - fac*cuCimag(tvxp1[ig]) );
        }
    }
}

extern "C"
void compute_excheng_(int *miwi, int *mjwi, double_complex *wfi_in, double_complex *wfj_in, double *exeng, double *exengs, double *wgti, double *wgtj) {
    cuDoubleComplex *wfi;
    int ig;
    int ib,jb;

    double fac;

    cufftResult stat;

    cuDoubleComplex *wf1;
    int miw, mjw;
    int ii, rbase;

    wfi = (cuDoubleComplex *) wfi_in;

    miw = *miwi;
    mjw = *mjwi;
    if (miw > niw || mjw > njw) {
        fprintf(stderr,"invalid miw, mjw = %d %d\n", miw, mjw);
        exit(1);
    }

    kernel_zero_d2 <<<nlx*nly, nlz>>> ((double2 *)ztmp);
    CUT_CHECK_ERROR("zero fail\n");
    kernel_zero_d2 <<<nlx*nly, nlz>>> ((double2 *)ztmp2);
    CUT_CHECK_ERROR("zero fail\n");

    kernel_zero_d <<<deviceProp.multiProcessorCount, 7>>> ((double *)SRmobj);
    CUT_CHECK_ERROR("zero fail\n");

    wf1 = wfi;
    for(ib=0; ib<miw; ib++) {
        CUDA_SAFE_CALL(cudaMemcpy(ztmp, wf1, nlxyz*sizeof(cuDoubleComplex), cudaMemcpyHostToDevice));
        wf1 = wf1 + nlxyz;

        stat = cufftExecZ2Z(plHandleB, ztmp, pxi, CUFFT_INVERSE);
        CHECKSTAT("fft bw %d\n",stat);

        kernel_zero_d2<<<nlx*nly,nlz>>>((double2 *)vsum);
        CUT_CHECK_ERROR("zero fail\n");

        for(jb=0; jb<mjw; jb++) {
            kernel_mul_zc_z<<<nlx*nly,nlz>>>((double2 *)(pxj[jb]),(double2 *)pxi,(double2 *)ztmp);
            CUT_CHECK_ERROR("mul_zc_z fail\n");

            /* */

            stat = cufftExecZ2Z(plHandleF, ztmp, ztmp2, CUFFT_FORWARD);
            CHECKSTAT("fft fw %d\n",stat);

            /* CKmobj includes FFT_FOWARD factor */

            kernel_zdscal3_v<<<nlx*nly,nlz>>>(CKmobj,(double2 *)ztmp2,vsum);
            CUT_CHECK_ERROR("zdscal_v fail\n");

            /* */

            stat = cufftExecZ2Z(plHandleB, vsum, ztmp, CUFFT_INVERSE);
            CHECKSTAT("fft bw %d\n",stat);

            /* */

            fac = -0.5*wgtj[jb]*wgti[ib]/nwxyz;
            rbase = 0;
            kernel_sumfzxy<<<deviceProp.multiProcessorCount, cores4sm, sizeof(double)*cores4sm>>>((double2 *)ztmp,(double2 *)pxj[jb],(double2 *)pxi,fac,nlxyz,SRmobj,rbase);
            CUT_CHECK_ERROR("sumfzxy fail");

            for(ii=0; ii<6; ii++) {
                /* CKSmobj includes FFT_FOWARD factor */

                kernel_zdscal3_v<<<nlx*nly,nlz>>>(CKSmobj[ii],(double2 *)ztmp2,vsum);
                CUT_CHECK_ERROR("zdscal_v fail\n");

                /* */

                stat = cufftExecZ2Z(plHandleB, vsum, ztmp, CUFFT_INVERSE);
                CHECKSTAT("fft bw %d\n",stat);

                /* */

                fac = -0.5*wgtj[jb]*wgti[ib]/nwxyz;
                rbase = deviceProp.multiProcessorCount*(ii+1);
                kernel_sumfzxy<<<deviceProp.multiProcessorCount,cores4sm,sizeof(double)*cores4sm>>>((double2 *)ztmp,(double2 *)pxj[jb],(double2 *)pxi,fac,nlxyz,SRmobj,rbase);
                CUT_CHECK_ERROR("sumfzxy fail");
            }
        }
    }
    CUDA_SAFE_CALL(cudaMemcpy(sumres, SRmobj, sizeof(double)*deviceProp.multiProcessorCount*7, cudaMemcpyDeviceToHost));

    for (ig=0; ig<deviceProp.multiProcessorCount; ig++) {
        *exeng = *exeng + sumres[ig];
    }
    for (ii=0; ii<6; ii++) {
        rbase = deviceProp.multiProcessorCount*(ii+1);
        for (ig=0; ig<deviceProp.multiProcessorCount; ig++) {
            exengs[ii] = exengs[ii] + sumres[rbase + ig];
        }
    }
}

/*
  Local Variables: ***
  mode:C ***
  End:***
*/
