!
! Copyright (c) 2000-2013,2017 Yoshihide Yoshimoto
!
#include "fftgrid.h"

module pot_gga_wb
  private
  public ggawb
contains
  subroutine ggawb(nspin,nrx2,nry2,nrz2,nlx,nly,nlz,nlxyz,nkfxa,kf, &
       rhop,rho,va,bb,exa,exc,evxc,excor,evxcor,fs,xctp,mssg,comm)
    !
    ! input
    ! rhop: partial core charge + valence charge (in real space)
    ! rho: valence charge (in real space)
    !
    ! output
    ! rho: Vxc(R)
    !
    use fft_3d
    use timeprof
    use mpi
    implicit none
    integer,intent(in):: nspin,nrx2,nry2,nrz2,nlx,nly,nlz,nlxyz
    integer,intent(in):: nkfxa,kf(3,nkfxa)
    !
    ! input rhop and rho must be padded by zero.
    !
    real(kind=8):: rhop(nlxyz,nspin)
    real(kind=8):: rho(nlxyz,nspin),va,bb(3,3),exa,exc,evxc(0:6)
    real(kind=8):: excor(nlxyz),evxcor(nlx*nly*nlz,0:6)
    type (fft3_struct):: fs
    integer,intent(in):: xctp
    integer,intent(in):: mssg, comm
    !
    real(kind=8):: minneg
    integer:: count
    real(kind=8),allocatable:: rtmp(:,:,:)
    integer:: is,ierr
    !
    ! make sure padding
    !
    do is=1,nspin
       call pad_zero(nrx2,nry2,nrz2,nlx,nly,nlz,rhop(1,is))
    end do
    !
    ! index 4 is working space.
    !
    allocate(FFT_A1M_DEFSZ(rtmp,nlxyz,2*nspin),stat=ierr)
    call chkalloc('ggawb','tmp',ierr)
    do is=1,nspin
       call gengrad(bb,nrx2,nry2,nrz2,nlx,nly,nlz,nkfxa,kf, &
            rhop(1,is),FFT_A1M_RL(rtmp,1,2*is-1),fs,comm)
    end do
    do is=1, nspin
       call fcpchr(nrx2,nry2,nrz2,nlx,nly,nlz,rhop(1,is),minneg,count)
       if (mssg.gt.0) then
          if (nspin .eq. 2) then
             write(6,'(1x,a,i1)') 'SPIN = ', is
          end if
          write(6,'(1x,a)') 'POSITIVE CHARGE FORCED ON RHOV + RHOP IN GGAWB'
          write(6,'(2x,i9,a,i9,a)') count,' of',nrx2*nry2*nrz2, &
               ' negative point found.'
          write(6,'(2x,a,e18.11)') 'Minimum value is ', minneg
          write(6,600) sum(rhop(:,is))*va/(nrx2*nry2*nrz2)
       end if
600    format(2x,'CORRECTED TOTAL CHARGE ',f25.20)
    end do
    if (nspin .eq. 1) then
       call evalxc(fs,nrx2,nry2,nrz2,nlx,nly,nlz,nlxyz,nkfxa,kf, &
            rhop,rtmp,rho,va,bb,exa,exc,evxc,excor,evxcor,xctp,mssg,comm)
    else if (nspin .eq. 2) then
       call evalxcsp(fs,nrx2,nry2,nrz2,nlx,nly,nlz,nlxyz,nkfxa,kf, &
            rhop,rtmp,rho,va,bb,exa,exc,evxc,excor,evxcor,xctp,mssg,comm)
    else
       write(6,*) 'invalid value in nspin = ', nspin
       stop
    end if
#ifdef INCONSISTENT_THREAD_COMP
    call tpf_begin(tpf_mpi)
    call MPI_BCAST(evxc,7,MPI_DOUBLE_PRECISION,0,comm,ierr)
    call MPI_BCAST(exa,1,MPI_DOUBLE_PRECISION,0,comm,ierr)
    call MPI_BCAST(exc,1,MPI_DOUBLE_PRECISION,0,comm,ierr)
    call tpf_end(tpf_mpi)
#endif
    deallocate(rtmp)
    return
  end subroutine ggawb

  subroutine fcpchr(nrx2,nry2,nrz2,nlx,nly,nlz,rho,minneg,count)
    implicit none
    integer,intent(in):: nlx,nly,nlz,nrx2,nry2,nrz2
    real(kind=8):: rho(nlx,nly,nlz),minneg
    integer:: count
    integer:: i,j,k
    minneg = rho(1,1,1)
    count = 0
    do k=1,nrz2
       do j=1,nry2
          do i=1,nrx2
             minneg = min(rho(i,j,k),minneg)
             if (rho(i,j,k) .le. 0.0d0) then
                rho(i,j,k) = 0.0d0
                count = count + 1
             end if
          end do
       end do
    end do
    return
  end subroutine fcpchr

  subroutine evalxc(fs,nrx2,nry2,nrz2,nlx,nly,nlz,nlxyz,nkfxa,kf, &
       rhop,drhop,rho,va,bb,exa,exc,evxc,excor,evxcor,xctp,mssg,comm)
    use fft_3d
    use pbe_xcgga
    use pbesol_xcgga
    use pw91_xcgga
    use pbe0_xcgga
    use b3lyp_xcgga
    use mpi
    use timeprof
    use xcf_types
    implicit none
    !
    ! input
    !
    type (fft3_struct):: fs
    integer,intent(in):: nrx2,nry2,nrz2,nlx,nly,nlz,nlxyz
    integer,intent(in):: nkfxa,kf(3,nkfxa)
    real(kind=8):: rhop(nlxyz),rho(nlxyz),va,bb(3,3)
    real(kind=8):: excor(nlxyz),evxcor(nlxyz,0:6)
    !
    ! index 4 is working space. (not used here)
    real(kind=8):: FFT_A1M_DEFSZ(drhop,nlxyz,2)
    integer,intent(in):: xctp
    integer,intent(in):: mssg,comm
    !
    ! output
    !
    ! rho contains Vxc
    real(kind=8):: exa,exc,evxc(0:6)
    !
    real(kind=8):: adreps
    parameter(adreps=1.0d-6)
    !
    integer:: k,j
    !
    real(kind=8),dimension(:),allocatable:: fxcn,fxcdn
    real(kind=8),allocatable:: pepsxc(:),pfxcn(:),pfxcdn(:),padrho(:)
    integer,allocatable:: r_count(:),r_displ(:)
    integer:: myrank,nproc,irnk,blxyz,pnlxyz
    real(kind=8):: pexc
    real(kind=8),allocatable:: ztmp(:,:)
    real(kind=8):: adrho
    integer:: ierr
    real(kind=8):: fx,fy,fz,rx,ry,rz,sevxc(1:6)
    !
    allocate(fxcdn(nlxyz),stat=ierr)
    call chkalloc('evalxc','tmp1',ierr)
    !
    call MPI_COMM_SIZE(comm,nproc,ierr)
    call MPI_COMM_RANK(comm,myrank,ierr)
    allocate(r_count(0:nproc-1),r_displ(0:nproc-1),stat=ierr)
    call chkalloc('evalxc','ptmp1',ierr)
    r_count = nlxyz/nproc
    do irnk = 0, mod(nlxyz,nproc)-1
       r_count(irnk) = r_count(irnk)+1
    end do
    r_displ(0) = 0
    do irnk = 0, nproc-2
       r_displ(irnk+1) = r_displ(irnk)+r_count(irnk)
    end do
    blxyz = r_displ(myrank)+1
    pnlxyz = r_count(myrank)
    allocate(padrho(pnlxyz),pepsxc(pnlxyz),pfxcn(pnlxyz), &
         pfxcdn(pnlxyz),stat=ierr)
    call chkalloc('evalxc','ptmp2',ierr)
    !
!$omp parallel do private(j)
    do k=1,pnlxyz
       j = r_displ(myrank)+k
       padrho(k) = sqrt( &
            FFT_A1M_RL(drhop,j,1)*FFT_A1M_RL(drhop,j,1) + FFT_A1M_IM(drhop,j,1)*FFT_A1M_IM(drhop,j,1) + &
            FFT_A1M_RL(drhop,j,2)*FFT_A1M_RL(drhop,j,2) )
    end do
    !
    select case(xctp)
    case(xcf_pbe)
       call pbe_exchcor(pnlxyz,rhop(blxyz),padrho, &
            pepsxc,pfxcn,pfxcdn,mssg)
    case(xcf_pbesol)
       call pbesol_exchcor(pnlxyz,rhop(blxyz),padrho, &
            pepsxc,pfxcn,pfxcdn,mssg)
    case(xcf_pw91)
       call pw91_exchcor(pnlxyz,rhop(blxyz),padrho, &
            pepsxc,pfxcn,pfxcdn,mssg)
    case(xcf_hartree_fock)
       if (mssg.gt.0) write(6,*) 'Hartree-Fock compensated by PBE'
       call pbe_exchcor(pnlxyz,rhop(blxyz),padrho, &
            pepsxc,pfxcn,pfxcdn,-1)
    case(xcf_pbe0)
       call pbe0_exchcor(pnlxyz,rhop(blxyz),padrho, &
            pepsxc,pfxcn,pfxcdn,mssg)
    case(xcf_pbe0_x)
       call pbe0_x_exchcor(pnlxyz,rhop(blxyz),padrho, &
            pepsxc,pfxcn,pfxcdn,mssg)
    case(xcf_b3lyp,xcf_b3lyp_loc)
       call b3lyp_exchcor(pnlxyz,rhop(blxyz),padrho, &
            pepsxc,pfxcn,pfxcdn,mssg)
    case(xcf_b3lyp_x)
       call b3lyp_x_exchcor(pnlxyz,rhop(blxyz),padrho, &
            pepsxc,pfxcn,pfxcdn,mssg)
    case default
       if (mssg.gt.0) then
          write(6,*) 'evalxc(): unsupported value in xctp = ', xctp
       end if
       call  MPI_ABORT(MPI_COMM_WORLD,1,ierr)
       stop
    end select
    !
    ! padding: at least rhop is zero
    !
    pexc = 0.0d0
    do k=1,pnlxyz
       pexc = pexc + rhop(r_displ(myrank)+k)*pepsxc(k)
    end do
    call tpf_begin(tpf_mpi)
    call MPI_ALLREDUCE(pexc,exc,1,MPI_DOUBLE_PRECISION,MPI_SUM, &
         comm,ierr)
    call MPI_ALLGATHERV(pfxcdn,pnlxyz,MPI_DOUBLE_PRECISION, &
         fxcdn,r_count,r_displ,MPI_DOUBLE_PRECISION,comm,ierr)
    !
    ! using pepsxc as a temporal array
    !
    do k=1,pnlxyz
       pepsxc(k) = rhop(r_displ(myrank)+k)*pepsxc(k)
    end do
    call MPI_ALLGATHERV(pepsxc,pnlxyz,MPI_DOUBLE_PRECISION, &
         excor,r_count,r_displ,MPI_DOUBLE_PRECISION,comm,ierr)
    call tpf_end(tpf_mpi)
    deallocate(padrho,pepsxc,pfxcdn)
    !
    ! make sure padding
    !
    call pad_zero(nrx2,nry2,nrz2,nlx,nly,nlz,fxcdn)
    !
    sevxc = 0.0d0
!$omp parallel do private(adrho,rx,ry,rz,fx,fy,fz), &
!$omp reduction (+:sevxc)
    do k=1,nlxyz
       adrho = sqrt( &
            FFT_A1M_RL(drhop,k,1)*FFT_A1M_RL(drhop,k,1) + FFT_A1M_IM(drhop,k,1)*FFT_A1M_IM(drhop,k,1) + &
            FFT_A1M_RL(drhop,k,2)*FFT_A1M_RL(drhop,k,2) )
       adrho = max(adrho,adreps)
       rx = FFT_A1M_RL(drhop,k,1)
       ry = FFT_A1M_IM(drhop,k,1)
       rz = FFT_A1M_RL(drhop,k,2)

       FFT_A1M_RL(drhop,k,1) = FFT_A1M_RL(drhop,k,1)/adrho*fxcdn(k)
       FFT_A1M_IM(drhop,k,1) = FFT_A1M_IM(drhop,k,1)/adrho*fxcdn(k)
       FFT_A1M_RL(drhop,k,2) = FFT_A1M_RL(drhop,k,2)/adrho*fxcdn(k)
       fx = FFT_A1M_RL(drhop,k,1)
       fy = FFT_A1M_IM(drhop,k,1)
       fz = FFT_A1M_RL(drhop,k,2)
       
       evxcor(k,1) = fx*rx
       evxcor(k,2) = fy*ry
       evxcor(k,3) = fz*rz
       evxcor(k,4) = fy*rz + fz*ry
       evxcor(k,5) = fx*rz + fz*rx
       evxcor(k,6) = fx*ry + fy*rx
       sevxc = sevxc + evxcor(k,1:6)
    end do
    evxc(1:6) = sevxc*va/(nrx2*nry2*nrz2)
    deallocate(fxcdn)
    allocate(FFT_A1D_DEFSZ(ztmp,nlxyz),stat=ierr)
    call chkalloc('evalxc','tmp2',ierr)
    call gendivr(bb,nrx2,nry2,nrz2,nlx,nly,nlz,nkfxa,kf, &
         drhop,ztmp,rhop,fs,comm)
    !
    allocate(fxcn(nlxyz),stat=ierr)
    call chkalloc('evalxc','tmp3',ierr)
    call tpf_begin(tpf_mpi)
    call MPI_ALLGATHERV(pfxcn,pnlxyz,MPI_DOUBLE_PRECISION, &
         fxcn,r_count,r_displ,MPI_DOUBLE_PRECISION,comm,ierr)
    call tpf_end(tpf_mpi)
    deallocate(pfxcn,r_count,r_displ)
    !
    call pad_zero(nrx2,nry2,nrz2,nlx,nly,nlz,fxcn)
    !
    ! generate Vxc(R) in rho

    call genvxc(nrx2*nry2*nrz2,nlxyz,fxcn,ztmp,rhop,rho,va,exa,exc,evxc,evxcor)
    deallocate(fxcn,ztmp)
    return
  end subroutine evalxc

  subroutine evalxcsp(fs,nrx2,nry2,nrz2,nlx,nly,nlz,nlxyz,nkfxa,kf, &
       rhop,drhop,rho,va,bb,exa,exc,evxc,excor,evxcor,xctp,mssg,comm)
    use fft_3d
    use pbe_xcgga
    use pbesol_xcgga
    use pw91_xcgga
    use pbe0_xcgga
    use b3lyp_xcgga
    use mpi
    use timeprof
    use xcf_types
    implicit none
    !
    ! input
    !
    type (fft3_struct):: fs
    integer,intent(in):: nrx2,nry2,nrz2,nlx,nly,nlz,nlxyz
    integer,intent(in):: nkfxa,kf(3,nkfxa)
    real(kind=8):: rhop(nlxyz,2),rho(nlxyz,2),va,bb(3,3)
    real(kind=8):: excor(nlxyz),evxcor(nlxyz,0:6)
    !
    ! index 4 is working space. (not used here)
    real(kind=8):: FFT_A1M_DEFSZ(drhop,nlxyz,4)
    integer,intent(in):: xctp
    integer,intent(in):: mssg,comm
    !
    ! output
    !
    !  rho contains Vxc
    real(kind=8):: exa,exc,evxc(0:6)
    !
    real(kind=8):: adreps
    parameter(adreps=1.0d-6)
    !
    integer:: k,j,is
    !
    real(kind=8),allocatable:: pfxcn(:,:),pfxdn(:,:)
    real(kind=8),allocatable:: fcdn(:)
    real(kind=8),allocatable:: prhop(:,:),padrho(:,:),padrhot(:)
    real(kind=8),allocatable:: ppfxcn(:,:),ppfxdn(:,:),pfcdn(:)
    real(kind=8),allocatable:: pexcor(:)
    integer,allocatable:: r_count(:),r_displ(:)
    integer:: myrank,nproc,irnk,pnlxyz
    real(kind=8),allocatable:: ztmp(:,:,:)
    integer:: ierr
    real(kind=8):: drhot(3), tfxdn(2),tfcdn
    real(kind=8):: pexc
    real(kind=8):: adrho(2),adrhot
    real(kind=8):: fx(3,2),rx(3,2),sevxc(1:6)
    !
    allocate(pfxdn(nlxyz,2),fcdn(nlxyz),stat=ierr)
    call chkalloc('evalxcsp','tmp1',ierr)
    !
    call MPI_COMM_SIZE(comm,nproc,ierr)
    call MPI_COMM_RANK(comm,myrank,ierr)
    allocate(r_count(0:nproc-1),r_displ(0:nproc-1),stat=ierr)
    call chkalloc('evalxcsp','ptmp1',ierr)
    r_count = nlxyz/nproc
    do irnk = 0, mod(nlxyz,nproc)-1
       r_count(irnk) = r_count(irnk)+1
    end do
    r_displ(0) = 0
    do irnk = 0, nproc-2
       r_displ(irnk+1) = r_displ(irnk)+r_count(irnk)
    end do
    pnlxyz = r_count(myrank)
    allocate(prhop(pnlxyz,2),padrho(pnlxyz,2),padrhot(pnlxyz), &
         ppfxcn(pnlxyz,2),ppfxdn(pnlxyz,2),pfcdn(pnlxyz), &
         pexcor(pnlxyz),stat=ierr)
    call chkalloc('evalxcsp','ptmp2',ierr)
    !
!$omp parallel do private(j,drhot)
    do k=1,pnlxyz
       j = r_displ(myrank)+k

       prhop(k,1) = rhop(j,1)
       prhop(k,2) = rhop(j,2)

       padrho(k,1) = sqrt(FFT_A1M_RL(drhop,j,1)*FFT_A1M_RL(drhop,j,1) &
            + FFT_A1M_IM(drhop,j,1)*FFT_A1M_IM(drhop,j,1) &
            + FFT_A1M_RL(drhop,j,2)*FFT_A1M_RL(drhop,j,2))
       padrho(k,2) = sqrt(FFT_A1M_RL(drhop,j,3)*FFT_A1M_RL(drhop,j,3) &
            + FFT_A1M_IM(drhop,j,3)*FFT_A1M_IM(drhop,j,3) &
            + FFT_A1M_RL(drhop,j,4)*FFT_A1M_RL(drhop,j,4))

       drhot(1) = FFT_A1M_RL(drhop,j,1) + FFT_A1M_RL(drhop,j,3)
       drhot(2) = FFT_A1M_IM(drhop,j,1) + FFT_A1M_IM(drhop,j,3)
       drhot(3) = FFT_A1M_RL(drhop,j,2) + FFT_A1M_RL(drhop,j,4)

       padrhot(k) = sqrt(drhot(1)*drhot(1) + drhot(2)*drhot(2) &
            + drhot(3)*drhot(3))
    end do
    select case(xctp)
    case (xcf_pbe)
       call pbe_exchcorsp(pnlxyz,prhop,padrho,padrhot, &
            pexc,pexcor,ppfxcn,ppfxdn,pfcdn,mssg)
    case (xcf_pbesol)
       call pbesol_exchcorsp(pnlxyz,prhop,padrho,padrhot, &
            pexc,pexcor,ppfxcn,ppfxdn,pfcdn,mssg)
    case (xcf_pw91)
       call pw91_exchcorsp(pnlxyz,prhop,padrho,padrhot, &
            pexc,pexcor,ppfxcn,ppfxdn,pfcdn,mssg)
    case (xcf_hartree_fock)
       if (mssg.gt.0) write(6,*) 'Hartree-Fock compensated by PBE'
       call pbe_exchcorsp(pnlxyz,prhop,padrho,padrhot, &
            pexc,pexcor,ppfxcn,ppfxdn,pfcdn,-1)
    case (xcf_pbe0)
       call pbe0_exchcorsp(pnlxyz,prhop,padrho,padrhot, &
            pexc,pexcor,ppfxcn,ppfxdn,pfcdn,mssg)
    case (xcf_pbe0_x)
       call pbe0_x_exchcorsp(pnlxyz,prhop,padrho,padrhot, &
            pexc,pexcor,ppfxcn,ppfxdn,pfcdn,mssg)
    case (xcf_b3lyp,xcf_b3lyp_loc)
       call b3lyp_exchcorsp(pnlxyz,prhop,padrho,padrhot, &
            pexc,pexcor,ppfxcn,ppfxdn,pfcdn,mssg)
    case (xcf_b3lyp_x)
       call b3lyp_x_exchcorsp(pnlxyz,prhop,padrho,padrhot, &
            pexc,pexcor,ppfxcn,ppfxdn,pfcdn,mssg)
    case default
       if (mssg.gt.0) then
          write(6,*) 'evalxcsp(): unsupported value in xctp = ', xctp
       end if
       call  MPI_ABORT(MPI_COMM_WORLD,1,ierr)
       stop
    end select
    !
    call tpf_begin(tpf_mpi)
    call MPI_ALLREDUCE(pexc,exc,1,MPI_DOUBLE_PRECISION,MPI_SUM, &
         comm,ierr)

    call MPI_ALLGATHERV(ppfxdn(1,1),pnlxyz,MPI_DOUBLE_PRECISION, &
         pfxdn(1,1),r_count,r_displ,MPI_DOUBLE_PRECISION,comm,ierr)
    call MPI_ALLGATHERV(ppfxdn(1,2),pnlxyz,MPI_DOUBLE_PRECISION, &
         pfxdn(1,2),r_count,r_displ,MPI_DOUBLE_PRECISION,comm,ierr)

    call MPI_ALLGATHERV(pfcdn,pnlxyz,MPI_DOUBLE_PRECISION, &
         fcdn,r_count,r_displ,MPI_DOUBLE_PRECISION,comm,ierr)

    call MPI_ALLGATHERV(pexcor,pnlxyz,MPI_DOUBLE_PRECISION, &
         excor,r_count,r_displ,MPI_DOUBLE_PRECISION,comm,ierr)
    call tpf_end(tpf_mpi)
    deallocate(ppfxdn,pfcdn,prhop,padrho,padrhot,pexcor)
    !
    do is=1,2 
       call pad_zero(nrx2,nry2,nrz2,nlx,nly,nlz,pfxdn(1,is))
    end do
    call pad_zero(nrx2,nry2,nrz2,nlx,nly,nlz,fcdn)
    !

    sevxc = 0.0d0
!$omp parallel do private(drhot,adrho,tfcdn,adrhot,tfxdn,fx,rx), &
!$omp reduction (+:sevxc)
    do k=1,nlxyz
       adrho(1) = sqrt(FFT_A1M_RL(drhop,k,1)*FFT_A1M_RL(drhop,k,1) &
            + FFT_A1M_IM(drhop,k,1)*FFT_A1M_IM(drhop,k,1) &
            + FFT_A1M_RL(drhop,k,2)*FFT_A1M_RL(drhop,k,2))
       adrho(2) = sqrt(FFT_A1M_RL(drhop,k,3)*FFT_A1M_RL(drhop,k,3) &
            + FFT_A1M_IM(drhop,k,3)*FFT_A1M_IM(drhop,k,3) &
            + FFT_A1M_RL(drhop,k,4)*FFT_A1M_RL(drhop,k,4))
       adrho = max(adrho,adreps)

       drhot(1) = FFT_A1M_RL(drhop,k,1) + FFT_A1M_RL(drhop,k,3)
       drhot(2) = FFT_A1M_IM(drhop,k,1) + FFT_A1M_IM(drhop,k,3)
       drhot(3) = FFT_A1M_RL(drhop,k,2) + FFT_A1M_RL(drhop,k,4)

       rx(1,1) = FFT_A1M_RL(drhop,k,1)
       rx(1,2) = FFT_A1M_RL(drhop,k,3)
       rx(2,1) = FFT_A1M_IM(drhop,k,1)
       rx(2,2) = FFT_A1M_IM(drhop,k,3)
       rx(3,1) = FFT_A1M_RL(drhop,k,2)
       rx(3,2) = FFT_A1M_RL(drhop,k,4)
       
       adrhot = sqrt(drhot(1)*drhot(1) + drhot(2)*drhot(2) &
            + drhot(3)*drhot(3))
       adrhot = max(adrhot,adreps)

       tfxdn(1) = FFT_A1M_RL(drhop,k,1)/adrho(1)*pfxdn(k,1)
       tfxdn(2) = FFT_A1M_RL(drhop,k,3)/adrho(2)*pfxdn(k,2)
       tfcdn = (FFT_A1M_RL(drhop,k,1)+FFT_A1M_RL(drhop,k,3))/adrhot*fcdn(k)
       FFT_A1M_RL(drhop,k,1) = tfxdn(1) + tfcdn
       FFT_A1M_RL(drhop,k,3) = tfxdn(2) + tfcdn

       tfxdn(1) = FFT_A1M_IM(drhop,k,1)/adrho(1)*pfxdn(k,1)
       tfxdn(2) = FFT_A1M_IM(drhop,k,3)/adrho(2)*pfxdn(k,2)
       tfcdn = (FFT_A1M_IM(drhop,k,1)+FFT_A1M_IM(drhop,k,3))/adrhot*fcdn(k)
       FFT_A1M_IM(drhop,k,1) = tfxdn(1) + tfcdn
       FFT_A1M_IM(drhop,k,3) = tfxdn(2) + tfcdn

       tfxdn(1) = FFT_A1M_RL(drhop,k,2)/adrho(1)*pfxdn(k,1)
       tfxdn(2) = FFT_A1M_RL(drhop,k,4)/adrho(2)*pfxdn(k,2)
       tfcdn = (FFT_A1M_RL(drhop,k,2)+FFT_A1M_RL(drhop,k,4))/adrhot*fcdn(k)
       FFT_A1M_RL(drhop,k,2) = tfxdn(1) + tfcdn
       FFT_A1M_RL(drhop,k,4) = tfxdn(2) + tfcdn

       fx(1,1) = FFT_A1M_RL(drhop,k,1)
       fx(1,2) = FFT_A1M_RL(drhop,k,3)
       fx(2,1) = FFT_A1M_IM(drhop,k,1)
       fx(2,2) = FFT_A1M_IM(drhop,k,3)
       fx(3,1) = FFT_A1M_RL(drhop,k,2)
       fx(3,2) = FFT_A1M_RL(drhop,k,4)

       evxcor(k,1) = fx(1,1)*rx(1,1) + fx(1,2)*rx(1,2)
       evxcor(k,2) = fx(2,1)*rx(2,1) + fx(2,2)*rx(2,2)
       evxcor(k,3) = fx(3,1)*rx(3,1) + fx(3,2)*rx(3,2)
       evxcor(k,4) = fx(2,1)*rx(3,1) + fx(3,1)*rx(2,1) &
            + fx(2,2)*rx(3,2) + fx(3,2)*rx(2,2)
       evxcor(k,5) = fx(1,1)*rx(3,1) + fx(3,1)*rx(1,1) &
            + fx(1,2)*rx(3,2) + fx(3,2)*rx(1,2)
       evxcor(k,6) = fx(1,1)*rx(2,1) + fx(2,1)*rx(1,1) &
            + fx(1,2)*rx(2,2) + fx(2,2)*rx(1,2)
       sevxc = sevxc + evxcor(k,1:6)
    end do
    deallocate(pfxdn,fcdn)
    evxc(1:6) = sevxc*va/(nrx2*nry2*nrz2)

    allocate(FFT_A1M_DEFSZ(ztmp,nlxyz,2),stat=ierr)
    call chkalloc('evalxcsp','tmp2',ierr)
    call gendivr(bb,nrx2,nry2,nrz2,nlx,nly,nlz,nkfxa,kf, &
         FFT_A1M_RL(drhop,1,1),FFT_A1M_RL(ztmp,1,1), &
         rhop(1,1),fs,comm)
    call gendivr(bb,nrx2,nry2,nrz2,nlx,nly,nlz,nkfxa,kf, &
         FFT_A1M_RL(drhop,1,3),FFT_A1M_RL(ztmp,1,2), &
         rhop(1,2),fs,comm)

    allocate(pfxcn(nlxyz,2),stat=ierr)
    call chkalloc('evalxcsp','tmp3',ierr)
    call tpf_begin(tpf_mpi)
    call MPI_ALLGATHERV(ppfxcn(1,1),pnlxyz,MPI_DOUBLE_PRECISION, &
         pfxcn(1,1),r_count,r_displ,MPI_DOUBLE_PRECISION,comm,ierr)
    call MPI_ALLGATHERV(ppfxcn(1,2),pnlxyz,MPI_DOUBLE_PRECISION, &
         pfxcn(1,2),r_count,r_displ,MPI_DOUBLE_PRECISION,comm,ierr)
    call tpf_end(tpf_mpi)
    deallocate(ppfxcn,r_count,r_displ)
    !
    do is=1,2
       call pad_zero(nrx2,nry2,nrz2,nlx,nly,nlz,pfxcn(1,is))
    end do

    ! generate Vxc(R) in rho
    call genvxcsp(nrx2*nry2*nrz2,nlxyz,pfxcn,ztmp,rhop,rho,va, &
         exa,exc,evxc,evxcor)
    deallocate(pfxcn,ztmp)
    return
  end subroutine evalxcsp
  !
  subroutine pad_zero(nx,ny,nz,lx,ly,lz,dat)
    implicit none
    integer,intent(in):: nx,ny,nz,lx,ly,lz
    real(kind=8):: dat(lx,ly,lz)

    integer:: j
    do j=nz+1,lz
       dat(:,:,j) = 0.0d0
    end do
    do j=nx+1,lx
       dat(j,:,:) = 0.0d0
    end do
    do j=ny+1,ly
       dat(:,j,:) = 0.0d0
    end do
    return
  end subroutine pad_zero

  subroutine gengrad(bb,nrx2,nry2,nrz2,nlx,nly,nlz,nkfxa,kf, &
       fun,grad,fs,comm)
    use fft_3d
    use mpi
    use timeprof
    implicit none
    !
    !     input
    !
    real(kind=8),intent(in):: bb(3,3)
    integer,intent(in):: nrx2,nry2,nrz2,nlx,nly,nlz
    integer,intent(in):: nkfxa,kf(3,nkfxa)
    real(kind=8),intent(in):: fun(nlx,nly,nlz)
    type (fft3_struct):: fs
    integer:: comm
    !
    !     output
    !
    real(kind=8),intent(out):: FFT_A3M_DEFSZ(grad,nlx,nly,nlz,2)
    !
    real(kind=8),allocatable:: work(:,:,:,:)
    !
    integer:: k,k1,k2,k3,j1,j2,j3,ierr
    real(kind=8):: fac1,fac2,fac3

    integer:: nlxyz2
    integer:: myrank,nproc,status(MPI_STATUS_SIZE)

    call MPI_COMM_SIZE(comm,nproc,ierr)
    call MPI_COMM_RANK(comm,myrank,ierr)

    allocate(FFT_A3D_DEFSZ(work,nlx,nly,nlz),stat=ierr)
    call chkalloc('gengrad','work',ierr)
    FFT_A3D_RL(work,:,:,:) = fun(:,:,:)
    FFT_A3D_IM(work,:,:,:) = 0.0d0
    call fft3_fw(fs,work,grad)
    !
    ! grad = I*G(j1,j2,j3)*work(j1,j2,j3)
    !
    grad = 0.0d0
!$omp parallel do private(k1,k2,k3,j1,j2,j3,fac1,fac2,fac3)
!OCL NOVREC(grad)
!POPTION INDEP(grad)
!VOPTION INDEP(grad)
!CDIR NODEP(grad)
!VOCL LOOP,DISJOINT(grad)
    do k=1,nkfxa
       k1=kf(1,k)
       k2=kf(2,k)
       k3=kf(3,k)
       j1=mod(nrx2+k1,nrx2)+1
       j2=mod(nry2+k2,nry2)+1
       j3=mod(nrz2+k3,nrz2)+1
       fac1 = bb(1,1)*k1 + bb(1,2)*k2 + bb(1,3)*k3
       fac2 = bb(2,1)*k1 + bb(2,2)*k2 + bb(2,3)*k3
       fac3 = bb(3,1)*k1 + bb(3,2)*k2 + bb(3,3)*k3
       FFT_A3M_RL(grad,j1,j2,j3,1) = - FFT_A3D_IM(work,j1,j2,j3)*fac1 - FFT_A3D_RL(work,j1,j2,j3)*fac2
       FFT_A3M_IM(grad,j1,j2,j3,1) = FFT_A3D_RL(work,j1,j2,j3)*fac1 - FFT_A3D_IM(work,j1,j2,j3)*fac2
       FFT_A3M_RL(grad,j1,j2,j3,2) = - FFT_A3D_IM(work,j1,j2,j3)*fac3
       FFT_A3M_IM(grad,j1,j2,j3,2) = FFT_A3D_RL(work,j1,j2,j3)*fac3
    end do

    if (nproc.gt.1) then
       nlxyz2 = nlx*nly*nlz*2
       if (mod(myrank,2).eq.0) then
          call fft3_bw(fs,FFT_A3M_RL(grad,1,1,1,1),work)
          call tpf_begin(tpf_mpi)
          if (myrank+1.lt.nproc) then
             call MPI_SENDRECV(FFT_A3M_RL(grad,1,1,1,1),nlxyz2, &
                  MPI_DOUBLE_PRECISION,myrank+1,0,FFT_A3M_RL(grad,1,1,1,2), &
                  nlxyz2,MPI_DOUBLE_PRECISION,myrank+1,0,comm,status,ierr)
          else
             call MPI_RECV(FFT_A3M_RL(grad,1,1,1,2),nlxyz2, &
                  MPI_DOUBLE_PRECISION,nproc-2,1,comm,status,ierr)
          end if
          call tpf_end(tpf_mpi)
       else
          call fft3_bw(fs,FFT_A3M_RL(grad,1,1,1,2),work)
          call tpf_begin(tpf_mpi)
          call MPI_SENDRECV(FFT_A3M_RL(grad,1,1,1,2),nlxyz2, &
               MPI_DOUBLE_PRECISION,myrank-1,0,FFT_A3M_RL(grad,1,1,1,1), &
               nlxyz2,MPI_DOUBLE_PRECISION,myrank-1,0,comm,status,ierr)
          if (mod(nproc,2).ne.0 .and. myrank.eq.nproc-2) then
             call MPI_SEND(FFT_A3M_RL(grad,1,1,1,2),nlxyz2, &
                  MPI_DOUBLE_PRECISION,nproc-1,1,comm,ierr)
          end if
          call tpf_end(tpf_mpi)
       end if
    else
       call fft3_bw(fs,FFT_A3M_RL(grad,1,1,1,1),work)
       call fft3_bw(fs,FFT_A3M_RL(grad,1,1,1,2),work)
    end if

    deallocate(work)
    return
  end subroutine gengrad
  !
  subroutine gendivr(bb,nrx2,nry2,nrz2,nlx,nly,nlz,nkfxa,kf, &
       vecf,divr,sclf,fs,comm)
    use fft_3d
    use mpi
    use timeprof
    implicit none
    !
    ! input
    !
    real(kind=8),intent(in):: bb(3,3)
    integer,intent(in):: nrx2,nry2,nrz2,nlx,nly,nlz
    integer,intent(in):: nkfxa,kf(3,nkfxa)
    !
    ! vecf will be destroyed at the exit
    real(kind=8):: FFT_A3M_DEFSZ(vecf,nlx,nly,nlz,2)
    real(kind=8):: sclf(nlx,nly,nlz)
    type (fft3_struct):: fs
    integer:: comm
    !
    ! output
    !
    real(kind=8):: FFT_A3D_DEFSZ(divr,nlx,nly,nlz)
    !
    integer:: k,k1,k2,k3,j1,j2,j3,i1,i2,i3
    real(kind=8):: g1,g2,g3
    real(kind=8):: fr1,fr2,fr3,fi1,fi2,fi3

    integer:: nlxyz2,ierr
    integer:: myrank,nproc,status(MPI_STATUS_SIZE)

    call MPI_COMM_SIZE(comm,nproc,ierr)
    call MPI_COMM_RANK(comm,myrank,ierr)

    if (nproc.gt.1) then
       nlxyz2 = nlx*nly*nlz*2
       if (mod(myrank,2).eq.0) then
          call fft3_fw(fs,FFT_A3M_RL(vecf,1,1,1,1),divr)
          call tpf_begin(tpf_mpi)
          if (myrank+1.lt.nproc) then
             call MPI_SENDRECV(FFT_A3M_RL(vecf,1,1,1,1),nlxyz2, &
                  MPI_DOUBLE_PRECISION,myrank+1,0,FFT_A3M_RL(vecf,1,1,1,2), &
                  nlxyz2,MPI_DOUBLE_PRECISION,myrank+1,0,comm,status,ierr)
          else
             call MPI_RECV(FFT_A3M_RL(vecf,1,1,1,2),nlxyz2, &
                  MPI_DOUBLE_PRECISION,nproc-2,1,comm,status,ierr)
          end if
          call tpf_end(tpf_mpi)
       else
          FFT_A3M_IM(vecf,:,:,:,2) = sclf
          call fft3_fw(fs,FFT_A3M_RL(vecf,1,1,1,2),divr)
          call tpf_begin(tpf_mpi)
          call MPI_SENDRECV(FFT_A3M_RL(vecf,1,1,1,2),nlxyz2, &
               MPI_DOUBLE_PRECISION,myrank-1,0,FFT_A3M_RL(vecf,1,1,1,1), &
               nlxyz2,MPI_DOUBLE_PRECISION,myrank-1,0,comm,status,ierr)
          if (mod(nproc,2).ne.0 .and. myrank.eq.nproc-2) then
             call MPI_SEND(FFT_A3M_RL(vecf,1,1,1,2),nlxyz2, &
                  MPI_DOUBLE_PRECISION,nproc-1,1,comm,ierr)
          end if
          call tpf_end(tpf_mpi)
       end if
    else
       call fft3_fw(fs,FFT_A3M_RL(vecf,1,1,1,1),divr)
       FFT_A3M_IM(vecf,:,:,:,2) = sclf
       call fft3_fw(fs,FFT_A3M_RL(vecf,1,1,1,2),divr)
    end if

    divr=0.0d0
!$omp parallel do private(k1,k2,k3,j1,j2,j3,i1,i2,i3,g1,g2,g3,fr1,fr2,fr3), &
!$omp private(fi1,fi2,fi3)
!OCL NOVREC(divr)
!POPTION INDEP(divr)
!VOPTION INDEP(divr)
!CDIR NODEP(divr)
!VOCL LOOP,DISJOINT(divr)
    do k=1,nkfxa
       k1=kf(1,k)
       k2=kf(2,k)
       k3=kf(3,k)
       j1=mod(nrx2+k1,nrx2)+1
       j2=mod(nry2+k2,nry2)+1
       j3=mod(nrz2+k3,nrz2)+1
       i1=mod(nrx2-k1,nrx2)+1
       i2=mod(nry2-k2,nry2)+1
       i3=mod(nrz2-k3,nrz2)+1
       !
       ! divr(j1,j2,j3) = I*<vecf(j1,j2,j3), G(j1,j2,j3)>
       !
       g1 = bb(1,1)*k1+bb(1,2)*k2+bb(1,3)*k3
       g2 = bb(2,1)*k1+bb(2,2)*k2+bb(2,3)*k3
       g3 = bb(3,1)*k1+bb(3,2)*k2+bb(3,3)*k3
       fr1 = 0.5d0*(  FFT_A3M_RL(vecf,j1,j2,j3,1) + FFT_A3M_RL(vecf,i1,i2,i3,1))
       fi1 = 0.5d0*(  FFT_A3M_IM(vecf,j1,j2,j3,1) - FFT_A3M_IM(vecf,i1,i2,i3,1))
       fr2 = 0.5d0*(  FFT_A3M_IM(vecf,j1,j2,j3,1) + FFT_A3M_IM(vecf,i1,i2,i3,1))
       fi2 = 0.5d0*(- FFT_A3M_RL(vecf,j1,j2,j3,1) + FFT_A3M_RL(vecf,i1,i2,i3,1))
       fr3 = 0.5d0*(  FFT_A3M_RL(vecf,j1,j2,j3,2) + FFT_A3M_RL(vecf,i1,i2,i3,2))
       fi3 = 0.5d0*(  FFT_A3M_IM(vecf,j1,j2,j3,2) - FFT_A3M_IM(vecf,i1,i2,i3,2))
       FFT_A3D_RL(divr,j1,j2,j3) = - fi1*g1 - fi2*g2 - fi3*g3
       FFT_A3D_IM(divr,j1,j2,j3) =   fr1*g1 + fr2*g2 + fr3*g3
    end do

    call fft3_bw(fs,divr,vecf)
    return
  end subroutine gendivr
  !
  subroutine genlap(bb,nrx2,nry2,nrz2,nlx,nly,nlz,nkfxa,kf, &
       fun,lap,fs)
    use fft_3d
    use mpi
    use timeprof
    implicit none
    !
    !     input
    !
    real(kind=8),intent(in):: bb(3,3)
    integer,intent(in):: nrx2,nry2,nrz2,nlx,nly,nlz
    integer,intent(in):: nkfxa,kf(3,nkfxa)
    real(kind=8),intent(in):: fun(nlx,nly,nlz)
    type (fft3_struct):: fs
    !
    !     output
    !
    real(kind=8),intent(out):: lap(nlx,nly,nlz)
    !
    real(kind=8),allocatable:: work(:,:,:,:),ztmp(:,:,:,:)
    !
    integer:: k,k1,k2,k3,j1,j2,j3,ierr
    real(kind=8):: fac1,fac2,fac3,fac

    allocate(FFT_A3D_DEFSZ(work,nlx,nly,nlz), &
         FFT_A3D_DEFSZ(ztmp,nlx,nly,nlz),stat=ierr)
    call chkalloc('genlap','work',ierr)
    FFT_A3D_RL(work,:,:,:) = fun(:,:,:)
    FFT_A3D_IM(work,:,:,:) = 0.0d0
    call fft3_fw(fs,work,ztmp)
    !
    ! ztmp = -G(j1,j2,j3)^2*work(j1,j2,j3)
    !
    ztmp = 0.0d0
!$omp parallel do private(k1,k2,k3,j1,j2,j3,fac1,fac2,fac3,fac)
!POPTION INDEP(ztmp)
!VOPTION INDEP(ztmp)
!CDIR NODEP(ztmp)
!VOCL LOOP,DISJOINT(ztmp)
    do k=1,nkfxa
       k1=kf(1,k)
       k2=kf(2,k)
       k3=kf(3,k)
       j1=mod(nrx2+k1,nrx2)+1
       j2=mod(nry2+k2,nry2)+1
       j3=mod(nrz2+k3,nrz2)+1
       fac1 = bb(1,1)*k1 + bb(1,2)*k2 + bb(1,3)*k3
       fac2 = bb(2,1)*k1 + bb(2,2)*k2 + bb(2,3)*k3
       fac3 = bb(3,1)*k1 + bb(3,2)*k2 + bb(3,3)*k3
       fac = -(fac1*fac1+fac2*fac2+fac3*fac3)
       FFT_A3D_RL(ztmp,j1,j2,j3) = fac*FFT_A3D_RL(work,j1,j2,j3)
       FFT_A3D_IM(ztmp,j1,j2,j3) = fac*FFT_A3D_IM(work,j1,j2,j3)
    end do

    call fft3_bw(fs,ztmp,work)
    lap(:,:,:) = FFT_A3D_RL(ztmp,:,:,:)

    deallocate(work,ztmp)
    return
  end subroutine genlap
  !
  subroutine genvxc(nrxyz,nlxyz,part1,part2,rhop,rho,va,exa,exc,evxc,evxcor)
    implicit none
    integer:: nrxyz,nlxyz
    real(kind=8):: part1(nlxyz),rhop(nlxyz),rho(nlxyz)
    real(kind=8):: FFT_A1D_DEFSZ(part2,nlxyz)
    real(kind=8):: va,exa,exc,evxc(0:6),evxcor(nlxyz,0:6)
    !
    integer:: i
    !
    evxc(0) = 0.0d0
    exa = exc
!$omp parallel do reduction(+:evxc,exa)
    do i=1,nlxyz
       evxc(0) = evxc(0) + part1(i)*rhop(i) - FFT_A1D_RL(part2,i)*rhop(i)
       evxcor(i,0) = part1(i)*rhop(i) - FFT_A1D_RL(part2,i)*rhop(i)
       exa = exa - part1(i)*rho(i) + FFT_A1D_RL(part2,i)*rho(i)
       rho(i) = part1(i) - FFT_A1D_RL(part2,i)
    end do
    exa = exa*va/dble(nrxyz)
    exc = exc*va/dble(nrxyz)
    evxc(0) = evxc(0)*va/dble(nrxyz)
    return
  end subroutine genvxc
  !
  subroutine genvxcsp(nrxyz,nlxyz,part1,part2,rhop,rho,va,exa,exc,evxc,evxcor)
    implicit none
    integer:: nrxyz,nlxyz
    real(kind=8):: part1(nlxyz,2),rhop(nlxyz,2),rho(nlxyz,2)
    real(kind=8):: FFT_A1M_DEFSZ(part2,nlxyz,2)
    real(kind=8):: va,exa,exc,evxc(0:6),evxcor(nlxyz,0:6)
    !
    integer:: i
    !
    evxc(0) = 0.0d0
    exa = exc
!$omp parallel do reduction(+:evxc,exa)
    do i=1,nlxyz
       evxc(0) = evxc(0) + rhop(i,1)*part1(i,1) + rhop(i,2)*part1(i,2) &
            - rhop(i,1)*FFT_A1M_RL(part2,i,1) - rhop(i,2)*FFT_A1M_RL(part2,i,2)
       evxcor(i,0) = rhop(i,1)*part1(i,1) + rhop(i,2)*part1(i,2) &
            - rhop(i,1)*FFT_A1M_RL(part2,i,1) - rhop(i,2)*FFT_A1M_RL(part2,i,2)
       exa = exa - rho(i,1)*( part1(i,1) - FFT_A1M_RL(part2,i,1) ) &
            - rho(i,2)*( part1(i,2) - FFT_A1M_RL(part2,i,2) )
       rho(i,1) = part1(i,1) - FFT_A1M_RL(part2,i,1)
       rho(i,2) = part1(i,2) - FFT_A1M_RL(part2,i,2)
    end do
    exa = exa*va/dble(nrxyz)
    exc = exc*va/dble(nrxyz)
    evxc(0) = evxc(0)*va/dble(nrxyz)
    return
  end subroutine genvxcsp
end module pot_gga_wb
