! Copyright (c) 1983-2000 TAPP Consortium
! Copyright (c) 2014 Yoshihide Yoshimoto
module subr_tetra
contains
  !
  ! tetra0(): integrate a list hexahedron determined by the points.
  ! A hexahedron is split into 6 tetrahedrons.
  ! The shape of each hexahedron must be the same parallelpiped.
  ! The position of the points are assumed to be
  !
  ! point 1: (0, 0, 0): ix,   iy,   iz
  ! point 2: (0, 1, 0): ix,   iy+1, iz
  ! point 3: (0, 0, 1): ix,   iy,   iz+1
  ! point 4: (0, 1, 1): ix,   iy+1, iz+1
  ! point 5: (1, 0, 0): ix+1, iy,   iz
  ! point 6: (1, 1, 0): ix+1, iy+1, iz
  ! point 7: (1, 0, 1): ix+1, iy,   iz+1
  ! point 8: (1, 1, 1): ix+1, iy+1, iz+1
  !
  ! Namely, the Jacobian must be considerd outside of this subroutine.
  !
  ! ef: fermi energy
  ! e*: eigen energy at points
  ! v*: integrand at points
  ! s: integral value
  !
  ! mode:
  !   0 : integral value up to ef
  !   1 : density = diff of integral at ef
  !
  ! This subroutine does not consider the spin degeneracy
  !
  subroutine tetra0(e1,e2,e3,e4,e5,e6,e7,e8 &
       ,v1,v2,v3,v4,v5,v6,v7,v8,ef,s,nis,mode)
    implicit none
    integer,intent(in):: nis, mode
    real(kind=8),intent(in):: ef
    real(kind=8):: s
    real(kind=8),intent(in):: e1(nis),e2(nis),e3(nis),e4(nis),e5(nis),e6(nis) &
         ,e7(nis),e8(nis)
    real(kind=8),intent(in):: v1(nis),v2(nis),v3(nis),v4(nis),v5(nis),v6(nis) &
         ,v7(nis),v8(nis)

    s = 0.0d0

    call tetra1(e1,e2,e3,e5,v1,v2,v3,v5,ef,s,nis,mode) 
    call tetra1(e2,e3,e5,e7,v2,v3,v5,v7,ef,s,nis,mode)
    call tetra1(e2,e5,e6,e7,v2,v5,v6,v7,ef,s,nis,mode)
    call tetra1(e4,e3,e2,e8,v4,v3,v2,v8,ef,s,nis,mode)
    call tetra1(e3,e2,e8,e6,v3,v2,v8,v6,ef,s,nis,mode)
    call tetra1(e3,e8,e7,e6,v3,v8,v7,v6,ef,s,nis,mode)
    
    return
  end subroutine tetra0

  subroutine tetra0c(e1,e2,e3,e4,e5,e6,e7,e8,ef,s,nis,mode)
    !
    ! v1 = v2 = v3 = v4 = 1
    !
    implicit none
    integer,intent(in):: nis, mode
    real(kind=8),intent(in):: ef
    real(kind=8):: s
    real(kind=8),intent(in):: e1(nis),e2(nis),e3(nis),e4(nis),e5(nis),e6(nis) &
         ,e7(nis),e8(nis)

    s = 0.0d0

    call tetra1c(e1,e2,e3,e5,ef,s,nis,mode) 
    call tetra1c(e2,e3,e5,e7,ef,s,nis,mode)
    call tetra1c(e2,e5,e6,e7,ef,s,nis,mode)
    call tetra1c(e4,e3,e2,e8,ef,s,nis,mode)
    call tetra1c(e3,e2,e8,e6,ef,s,nis,mode)
    call tetra1c(e3,e8,e7,e6,ef,s,nis,mode)
    
    return
  end subroutine tetra0c

  subroutine tetra1(e1,e2,e3,e4,v1,v2,v3,v4,ef,s,nis,mode)
    implicit none
    integer,intent(in):: nis,mode
    real(kind=8),intent(in):: ef
    real(kind=8):: s
    real(kind=8),intent(in):: e1(nis),e2(nis),e3(nis),e4(nis)
    real(kind=8),intent(in):: v1(nis),v2(nis),v3(nis),v4(nis)

    integer:: in
    real(kind=8):: ee1,ee2,ee3,ee4
    real(kind=8):: vv1,vv2,vv3,vv4
    real(kind=8):: a1,b1,a2,b2,c1,d1,c2,d2

    do in=1,nis
       if(e1(in).ge.e2(in)) then
          a1=e1(in)
          b1=v1(in)
          a2=e2(in)
          b2=v2(in)
       else
          a1=e2(in)
          b1=v2(in)
          a2=e1(in)
          b2=v1(in)
       end if

       if(e3(in).ge.e4(in)) then
          c1=e3(in)
          d1=v3(in)
          c2=e4(in)
          d2=v4(in)
       else
          c1=e4(in)
          d1=v4(in)
          c2=e3(in)
          d2=v3(in)
       end if

       if(a1.ge.c1) then
          ee1=a1
          vv1=b1
          a1=c1
          b1=d1
       else
          ee1=c1
          vv1=d1
       end if

       if(a2.ge.c2) then
          ee4=c2
          vv4=d2
       else
          ee4=a2
          vv4=b2
          a2=c2
          b2=d2
       end if

       if(a1.ge.a2) then
          ee2=a1
          vv2=b1
          ee3=a2
          vv3=b2
       else
          ee2=a2
          vv2=b2
          ee3=a1
          vv3=b1
       end if

       if (mode.eq.0) then
          call tetra2(ee1,ee2,ee3,ee4,vv1,vv2,vv3,vv4,ef,s)
       else
          call tetra2d(ee1,ee2,ee3,ee4,vv1,vv2,vv3,vv4,ef,s)
       end if
    end do

    return
  end subroutine tetra1

  subroutine tetra1c(e1,e2,e3,e4,ef,s,nis,mode)
    !
    ! v1 = v2 = v3 = v4 = 1
    !
    implicit none
    integer,intent(in):: nis,mode
    real(kind=8),intent(in):: ef
    real(kind=8):: s
    real(kind=8),intent(in):: e1(nis),e2(nis),e3(nis),e4(nis)

    integer:: in
    real(kind=8):: ee1,ee2,ee3,ee4
    real(kind=8):: a1,a2,c1,c2

    do in=1,nis
       if(e1(in).ge.e2(in)) then
          a1=e1(in)
          a2=e2(in)
       else
          a1=e2(in)
          a2=e1(in)
       end if

       if(e3(in).ge.e4(in)) then
          c1=e3(in)
          c2=e4(in)
       else
          c1=e4(in)
          c2=e3(in)
       end if

       if(a1.ge.c1) then
          ee1=a1
          a1=c1
       else
          ee1=c1
       end if

       if(a2.ge.c2) then
          ee4=c2
       else
          ee4=a2
          a2=c2
       end if

       if(a1.ge.a2) then
          ee2=a1
          ee3=a2
       else
          ee2=a2
          ee3=a1
       end if

       if (mode.eq.0) then
          call tetra2c(ee1,ee2,ee3,ee4,ef,s)
       else
          call tetra2cd(ee1,ee2,ee3,ee4,ef,s)
       end if
    end do

    return
  end subroutine tetra1c

  subroutine tetra2(e1,e2,e3,e4,v1,v2,v3,v4,ef,s)
    implicit none
    real(kind=8),intent(in):: ef
    real(kind=8):: s
    real(kind=8),intent(in):: e1,e2,e3,e4
    real(kind=8),intent(in):: v1,v2,v3,v4

    logical:: lb

    ! spin is not considered
    real(kind=8),parameter:: unit = 1.0d0/24.0d0

    if((e1.eq.e2).and.(ef.eq.e2)) then
       lb=.true.
    else
       lb=.false.
    end if

    if (ef.ge.e1) then
       s=s+(v1+v2+v3+v4)*unit
    else if (ef.gt.e2) then
       s=s+(v1+v2+v3+v4)*unit &
            -unit*(e1-ef)*(e1-ef)*(e1-ef) &
            /(e1-e4)/(e1-e3)/(e1-e2) &
            *(4.0d0*v1+(ef-e1)*((v4-v1)/(e4-e1) &
            +(v3-v1)/(e3-e1) &
            +(v2-v1)/(e2-e1)))
    else if ((ef.ge.e3).and.(ef.le.e2).and.(e2.ne.e3).and. &
         (.not.lb)) then
       s=s+unit*((ef-e1)*(ef-e4)*(ef-e4) &
            /(e3-e1)/(e2-e4)/(e1-e4) &
            *(3.0d0*v4+(ef-e4)*(v2-v4)/(e2-e4) &
            +v3+(ef-e4)*(v1-v4)/(e1-e4) &
            +(ef-e3)*(v1-v3)/(e1-e3)) &
            +(ef-e4)*(e2-ef)*(ef-e3) &
            /(e2-e4)/(e2-e3)/(e1-e3) &
            *(2.0d0*v4+(ef-e4)*(v2-v4)/(e2-e4) &
            +2.0d0*v3+(ef-e3)*(v2-v3)/(e2-e3) &
            +(ef-e3)*(v1-v3)/(e1-e3)) &
            +(ef-e3)*(ef-e3)/(e1-e3)/(e2-e3) &
            *(3.0d0*v3+(ef-e3)*(v2-v3)/(e2-e3) &
            +v4+(ef-e3)*(v1-v3)/(e1-e3)))
    else if ((ef.eq.e3).and.(e3.eq.e2).and.(.not.lb)) then
       s=s+unit*(e3-e4)/(e1-e4) &
            *(v2+(e3-e4)*(v1-v4)/(e1-e4) &
            +v3+2.0d0*v4)
    else if ((ef.gt.e4).and.(ef.lt.e3)) then
       s=s-unit*(e4-ef)*(e4-ef)*(e4-ef) &
            /(e1-e4)/(e2-e4)/(e3-e4) &
            *(4.0d0*v4+(ef-e4)*((v3-v4)/(e3-e4) &
            +(v1-v4)/(e1-e4) &
            +(v2-v4)/(e2-e4)))
    end if

    return
  end subroutine tetra2

  subroutine tetra2d(e1,e2,e3,e4,v1,v2,v3,v4,ef,s)
    !
    ! derivative
    !
    implicit none
    real(kind=8),intent(in):: ef
    real(kind=8):: s
    real(kind=8),intent(in):: e1,e2,e3,e4
    real(kind=8),intent(in):: v1,v2,v3,v4

    logical:: lb

    ! spin is not considered
    real(kind=8),parameter:: unit = 1.0d0/24.0d0

    if((e1.eq.e2).and.(ef.eq.e2)) then
       lb=.true.
    else
       lb=.false.
    end if

    if (ef.ge.e1) then
       continue
    else if (ef.gt.e2) then
       s = s + 3*(e1-ef)**2*unit*((ef-e1)*((v4-v1)/(e4-e1)+(v3-v1)/(e3-e1)+(v2-v1)/(e2-e1))+4*v1)/((e1-e2)*(e1-e3)*(e1-e4)) &
            -(e1-ef)**3*unit*((v4-v1)/(e4-e1)+(v3-v1)/(e3-e1)+(v2-v1)/(e2-e1))/((e1-e2)*(e1-e3)*(e1-e4))
    else if ((ef.ge.e3).and.(ef.le.e2).and.(e2.ne.e3).and. &
         (.not.lb)) then
       s = s + unit*((ef-e4)**2*(3*v4+(ef-e4)*(v2-v4)/(e2-e4) &
            +(ef-e4)*(v1-v4)/(e1-e4)+v3+(ef-e3)*(v1-v3)/(e1-e3))/((e3-e1)*(e1-e4)*(e2-e4)) &
            +2*(ef-e1)*(ef-e4)*(3*v4+(ef-e4)*(v2-v4)/(e2-e4)+(ef-e4)*(v1-v4)/(e1-e4) &
            +v3+(ef-e3)*(v1-v3)/(e1-e3))/((e3-e1)*(e1-e4)*(e2-e4)) &
            -(ef-e3)*(ef-e4)*(2*v4+(ef-e4)*(v2-v4)/(e2-e4)+2*v3+(ef-e3)*(v2-v3)/(e2-e3) &
            +(ef-e3)*(v1-v3)/(e1-e3))/((e1-e3)*(e2-e3)*(e2-e4)) &
            +(e2-ef)*(ef-e4)*(2*v4+(ef-e4)*(v2-v4)/(e2-e4)+2*v3+(ef-e3)*(v2-v3)/(e2-e3) &
            +(ef-e3)*(v1-v3)/(e1-e3))/((e1-e3)*(e2-e3)*(e2-e4)) &
            +(e2-ef)*(ef-e3)*(2*v4+(ef-e4)*(v2-v4)/(e2-e4)+2*v3+(ef-e3)*(v2-v3)/(e2-e3) &
            +(ef-e3)*(v1-v3)/(e1-e3))/((e1-e3)*(e2-e3)*(e2-e4)) &
            +2*(ef-e3)*(v4+3*v3+(ef-e3)*(v2-v3)/(e2-e3)+(ef-e3)*(v1-v3)/(e1-e3))/((e1-e3)*(e2-e3)) &
            +(ef-e1)*(ef-e4)**2*((v2-v4)/(e2-e4) &
            +(v1-v4)/(e1-e4)+(v1-v3)/(e1-e3))/((e3-e1)*(e1-e4)*(e2-e4)) &
            +(e2-ef)*(ef-e3)*(ef-e4)*((v2-v4)/(e2-e4)+(v2-v3)/(e2-e3)+(v1-v3)/(e1-e3))/((e1-e3)*(e2-e3)*(e2-e4)) &
            +(ef-e3)**2*((v2-v3)/(e2-e3)+(v1-v3)/(e1-e3))/((e1-e3)*(e2-e3)))
    else if ((ef.eq.e3).and.(e3.eq.e2).and.(.not.lb)) then
       continue
    else if ((ef.gt.e4).and.(ef.lt.e3)) then
       s = s + 3*(e4-ef)**2*unit*(4*v4+(ef-e4)*((v3-v4)/(e3-e4)+(v2-v4)/(e2-e4)+(v1-v4)/(e1-e4)))/((e1-e4)*(e2-e4)*(e3-e4)) &
            -(e4-ef)**3*unit*((v3-v4)/(e3-e4)+(v2-v4)/(e2-e4)+(v1-v4)/(e1-e4))/((e1-e4)*(e2-e4)*(e3-e4))
    end if

    return
  end subroutine tetra2d

  subroutine tetra2c(e1,e2,e3,e4,ef,s)
    !
    ! v1 = v2 = v3 = v4 = 1
    !
    implicit none
    real(kind=8),intent(in):: ef
    real(kind=8):: s
    real(kind=8),intent(in):: e1,e2,e3,e4

    logical:: lb

    ! spin is not considered
    real(kind=8),parameter:: unit = 1.0d0/24.0d0

    if((e1.eq.e2).and.(ef.eq.e2)) then
       lb=.true.
    else
       lb=.false.
    end if

    if (ef.ge.e1) then
       s=s+4*unit
    else if (ef.gt.e2) then
       s=s+4*unit-4*(e1-ef)**3*unit/((e1-e2)*(e1-e3)*(e1-e4))
    else if ((ef.ge.e3).and.(ef.le.e2).and.(e2.ne.e3).and. &
         (.not.lb)) then
       s=s+(4*(ef-e1)*(ef-e4)**2/((e3-e1)*(e1-e4)*(e2-e4)) &
            +4*(ef-e3)**2/((e1-e3)*(e2-e3)) &
            +4*(e2-ef)*(ef-e3)*(ef-e4)/((e1-e3)*(e2-e3)*(e2-e4)))*unit
    else if ((ef.eq.e3).and.(e3.eq.e2).and.(.not.lb)) then
       s=s+4*(e3-e4)*unit/(e1-e4)
    else if ((ef.gt.e4).and.(ef.lt.e3)) then
       s=s-4*(e4-ef)**3*unit/((e1-e4)*(e2-e4)*(e3-e4))
    end if

    return
  end subroutine tetra2c

  subroutine tetra2cd(e1,e2,e3,e4,ef,s)
    !
    ! v1 = v2 = v3 = v4 = 1
    ! derivative
    !
    implicit none
    real(kind=8),intent(in):: ef
    real(kind=8):: s
    real(kind=8),intent(in):: e1,e2,e3,e4

    logical:: lb

    ! spin is not considered
    real(kind=8),parameter:: unit = 1.0d0/24.0d0

    if((e1.eq.e2).and.(ef.eq.e2)) then
       lb=.true.
    else
       lb=.false.
    end if

    if (ef.ge.e1) then
       continue
    else if (ef.gt.e2) then
       s=s+12*(e1-ef)**2*unit/((e1-e2)*(e1-e3)*(e1-e4))
    else if ((ef.ge.e3).and.(ef.le.e2).and.(e2.ne.e3).and. &
         (.not.lb)) then
       s=s+(4*(ef-e4)**2/((e3-e1)*(e1-e4)*(e2-e4)) &
            -4*(ef-e3)*(ef-e4)/((e1-e3)*(e2-e3)*(e2-e4)) &
            +8*(ef-e1)*(ef-e4)/((e3-e1)*(e1-e4)*(e2-e4)) &
            +4*(e2-ef)*(ef-e4)/((e1-e3)*(e2-e3)*(e2-e4)) &
            +4*(e2-ef)*(ef-e3)/((e1-e3)*(e2-e3)*(e2-e4)) &
            +8*(ef-e3)/((e1-e3)*(e2-e3)))*unit
    else if ((ef.eq.e3).and.(e3.eq.e2).and.(.not.lb)) then
       continue
    else if ((ef.gt.e4).and.(ef.lt.e3)) then
       s=s+12*(e4-ef)**2*unit/((e1-e4)*(e2-e4)*(e3-e4))
    end if

    return
  end subroutine tetra2cd
end module subr_tetra
